from __future__ import annotations

import json
import logging
import os
from typing import Any

from langchain_openai import ChatOpenAI

from schemas import SubQuestionAnswer

logger = logging.getLogger(__name__)

_REFINEMENT_DECOMPOSITION_MODEL = os.getenv("REFINEMENT_DECOMPOSITION_MODEL", "gpt-4.1-mini")
_REFINEMENT_DECOMPOSITION_TEMPERATURE = float(os.getenv("REFINEMENT_DECOMPOSITION_TEMPERATURE", "0"))
_REFINEMENT_DECOMPOSITION_MAX_SUBQUESTIONS = max(1, int(os.getenv("REFINEMENT_DECOMPOSITION_MAX_SUBQUESTIONS", "6")))
_OPENAI_API_KEY = os.getenv("OPENAI_API_KEY", "").strip()


def _normalize_question(text: str) -> str:
    normalized = (text or "").strip()
    if not normalized:
        return ""
    normalized = normalized.rstrip("?.! ").strip()
    if not normalized:
        return ""
    return f"{normalized}?"


def _format_sub_qa(sub_qa: list[SubQuestionAnswer]) -> str:
    lines: list[str] = []
    for index, item in enumerate(sub_qa, start=1):
        lines.append(
            "\n".join(
                [
                    f"[{index}] sub_question={item.sub_question}",
                    f"answerable={item.answerable}",
                    f"verification_reason={item.verification_reason}",
                    f"sub_answer={item.sub_answer}",
                ]
            )
        )
    return "\n\n".join(lines)


def _sanitize_refined_subquestions(
    *,
    candidates: list[str],
    existing_subquestions: list[str],
) -> list[str]:
    existing_normalized = {
        _normalize_question(question).lower()
        for question in existing_subquestions
        if _normalize_question(question)
    }
    output: list[str] = []
    seen: set[str] = set()

    for candidate in candidates:
        normalized = _normalize_question(candidate)
        if not normalized:
            continue
        lowered = normalized.lower()
        if lowered in existing_normalized or lowered in seen:
            continue
        output.append(normalized)
        seen.add(lowered)
        if len(output) >= _REFINEMENT_DECOMPOSITION_MAX_SUBQUESTIONS:
            break
    return output


def _extract_llm_subquestions(content: Any) -> list[str]:
    if isinstance(content, list):
        return [str(item).strip() for item in content if str(item).strip()]

    text = str(content or "").strip()
    if not text:
        return []

    try:
        parsed = json.loads(text)
        if isinstance(parsed, list):
            return [str(item).strip() for item in parsed if str(item).strip()]
    except json.JSONDecodeError:
        pass

    lines = [line.strip("-* \t") for line in text.splitlines() if line.strip()]
    return [line for line in lines if line]


def _fallback_refined_subquestions(
    *,
    question: str,
    initial_answer: str,
    sub_qa: list[SubQuestionAnswer],
) -> list[str]:
    _ = initial_answer
    unresolved = [item for item in sub_qa if not item.answerable]
    candidates: list[str] = []

    for item in unresolved:
        normalized_subq = _normalize_question(item.sub_question)
        if not normalized_subq:
            continue
        candidates.append(f"What specific missing evidence is needed to answer {normalized_subq.rstrip('?')}?")
        if item.verification_reason:
            reason = item.verification_reason.replace("_", " ").strip()
            candidates.append(f"What sources can resolve this issue: {reason}?")

    if not candidates:
        candidates.append(
            f"What additional evidence is needed to fully answer {_normalize_question(question).rstrip('?')}?"
        )

    return _sanitize_refined_subquestions(
        candidates=candidates,
        existing_subquestions=[item.sub_question for item in sub_qa],
    )


def refine_subquestions(
    *,
    question: str,
    initial_answer: str,
    sub_qa: list[SubQuestionAnswer],
    callbacks: list[Any] | None = None,
) -> list[str]:
    logger.info(
        "Refinement decomposition start question_len=%s initial_answer_len=%s sub_qa_count=%s",
        len(question),
        len(initial_answer),
        len(sub_qa),
    )

    if not _OPENAI_API_KEY:
        fallback = _fallback_refined_subquestions(
            question=question,
            initial_answer=initial_answer,
            sub_qa=sub_qa,
        )
        logger.info(
            "Refinement decomposition complete via fallback count=%s",
            len(fallback),
        )
        return fallback

    prompt = (
        "You generate refined sub-questions for a retrieval pipeline.\n"
        "Goal: target gaps in the initial answer and unresolved sub-questions.\n\n"
        "Rules:\n"
        "- Output only refined sub-questions.\n"
        "- Do not repeat existing sub-questions.\n"
        "- Each output must be a complete question ending with '?'.\n"
        "- Keep each sub-question narrow and specific.\n"
        f"- Return at most {_REFINEMENT_DECOMPOSITION_MAX_SUBQUESTIONS} questions.\n"
        "- Return valid JSON as an array of strings.\n\n"
        f"User question:\n{question}\n\n"
        f"Initial answer:\n{initial_answer}\n\n"
        f"Existing sub-question results:\n{_format_sub_qa(sub_qa) or 'None'}\n"
    )

    try:
        llm = ChatOpenAI(
            model=_REFINEMENT_DECOMPOSITION_MODEL,
            temperature=_REFINEMENT_DECOMPOSITION_TEMPERATURE,
        )
        invoke_config = {"callbacks": callbacks} if callbacks else None
        response = llm.invoke(prompt, config=invoke_config) if invoke_config else llm.invoke(prompt)
        raw_candidates = _extract_llm_subquestions(getattr(response, "content", ""))
        refined = _sanitize_refined_subquestions(
            candidates=raw_candidates,
            existing_subquestions=[item.sub_question for item in sub_qa],
        )
        if refined:
            logger.info(
                "Refinement decomposition complete via LLM count=%s model=%s",
                len(refined),
                _REFINEMENT_DECOMPOSITION_MODEL,
            )
            return refined
        logger.warning("Refinement decomposition returned no valid LLM sub-questions; using fallback")
    except Exception:
        logger.exception(
            "Refinement decomposition LLM call failed; using fallback model=%s",
            _REFINEMENT_DECOMPOSITION_MODEL,
        )

    fallback = _fallback_refined_subquestions(
        question=question,
        initial_answer=initial_answer,
        sub_qa=sub_qa,
    )
    logger.info(
        "Refinement decomposition complete via fallback count=%s",
        len(fallback),
    )
    return fallback



from agent_search.runtime.execution_identity import (
    ExecutionIdentityError,
    mint_thread_id,
    resolve_thread_identity,
    validate_thread_id,
)
from agent_search.runtime.resume import ResumeTransitionError
from agent_search.runtime.state import RAGState

__all__ = [
    "ExecutionIdentityError",
    "RAGState",
    "ResumeTransitionError",
    "mint_thread_id",
    "resolve_thread_identity",
    "run_runtime_agent",
    "validate_thread_id",
]


def __getattr__(name: str):
    if name == "run_runtime_agent":
        from agent_search.runtime.runner import run_runtime_agent

        return run_runtime_agent
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



__all__ = ["advanced_rag"]


def __getattr__(name: str):
    if name == "advanced_rag":
        from . import public_api

        return getattr(public_api, name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from __future__ import annotations

import inspect
import logging
from typing import Any

from langchain_core.documents import Document

logger = logging.getLogger(__name__)


class LangChainVectorStoreAdapter:
    """Adapter exposing a stable retrieval interface over LangChain vector stores."""

    def __init__(self, store: Any) -> None:
        self._store = store

    def similarity_search(
        self,
        query: str,
        k: int,
        filter: dict[str, Any] | None = None,
    ) -> list[Document]:
        safe_k = max(1, k)
        method = getattr(self._store, "similarity_search", None)
        if not callable(method):
            raise TypeError("Wrapped store must implement callable similarity_search")

        try:
            documents = method(query, k=safe_k, filter=filter)
            logger.info(
                "LangChain adapter similarity_search query_len=%s k=%s has_filter=%s results=%s mode=with_filter",
                len(query),
                safe_k,
                filter is not None,
                len(documents),
            )
            return documents
        except TypeError:
            documents = method(query, k=safe_k)
            logger.info(
                "LangChain adapter similarity_search fallback query_len=%s k=%s has_filter=%s results=%s mode=without_filter",
                len(query),
                safe_k,
                filter is not None,
                len(documents),
            )
            return documents

    def similarity_search_with_relevance_scores(
        self,
        query: str,
        k: int,
        score_threshold: float | None = None,
        filter: dict[str, Any] | None = None,
    ) -> list[tuple[Document, float]]:
        safe_k = max(1, k)
        method = getattr(self._store, "similarity_search_with_relevance_scores", None)

        if callable(method):
            kwargs: dict[str, Any] = {"k": safe_k}
            if score_threshold is not None:
                kwargs["score_threshold"] = score_threshold
            if filter is not None:
                try:
                    signature = inspect.signature(method)
                except (TypeError, ValueError):
                    signature = None
                if signature is None or "filter" in signature.parameters:
                    kwargs["filter"] = filter

            docs_with_scores = method(query, **kwargs)
            logger.info(
                "LangChain adapter relevance_search query_len=%s k=%s score_threshold=%s has_filter=%s results=%s mode=with_scores",
                len(query),
                safe_k,
                score_threshold,
                filter is not None,
                len(docs_with_scores),
            )
            return docs_with_scores

        documents = self.similarity_search(query, k=safe_k, filter=filter)
        logger.info(
            "LangChain adapter relevance_search fallback query_len=%s k=%s score_threshold=%s has_filter=%s results=%s mode=similarity_search",
            len(query),
            safe_k,
            score_threshold,
            filter is not None,
            len(documents),
        )
        return [(document, 1.0) for document in documents]




from agent_search.runtime.nodes.answer import run_answer_node
from agent_search.runtime.nodes.decompose import run_decomposition_node
from agent_search.runtime.nodes.expand import run_expansion_node
from agent_search.runtime.nodes.rerank import run_rerank_node
from agent_search.runtime.nodes.search import run_search_node
from agent_search.runtime.nodes.synthesize import run_synthesize_node

__all__ = [
    "run_answer_node",
    "run_decomposition_node",
    "run_expansion_node",
    "run_search_node",
    "run_rerank_node",
    "run_synthesize_node",
]



from __future__ import annotations

from concurrent.futures import TimeoutError as FuturesTimeoutError
import logging
from typing import Any, Mapping

from agent_search.config import RuntimeConfig
from agent_search.errors import (
    SDKConfigurationError,
    SDKError,
    SDKModelError,
    SDKRetrievalError,
    SDKTimeoutError,
)
from agent_search.runtime.graph.builder import build_runtime_graph
from agent_search.runtime.graph.state import RuntimeGraphContext, to_runtime_graph_state
from agent_search.runtime.persistence import compile_graph_with_checkpointer
from agent_search.runtime.resume import build_resume_command
from agent_search.runtime.runner import run_runtime_agent
from agent_search.vectorstore.protocol import assert_vector_store_compatible
from pydantic import ValidationError
from schemas import (
    RuntimeAgentRunControls,
    RuntimeAgentRunRequest,
    RuntimeAgentRunResponse,
    RuntimeHitlControl,
    RuntimeQueryExpansionHitlControl,
    RuntimeSubquestionHitlControl,
)
from schemas.agent import (
    HitlResumeRequest,
    HitlReview,
    RuntimeAgentRunResult,
    RuntimeQueryExpansionDecision,
    RuntimeQueryExpansionResumeEnvelope,
    RuntimeSubquestionDecision,
    RuntimeSubquestionResumeEnvelope,
)
from services import agent_service as legacy_service

logger = logging.getLogger(__name__)
_CUSTOM_PROMPT_KEYS = ("subanswer", "synthesis")


def _resolve_request_thread_id(config: Mapping[str, Any] | None = None) -> str | None:
    if not isinstance(config, Mapping):
        return None
    thread_id = config.get("thread_id")
    if thread_id is None:
        return None
    return str(thread_id)


def _resolve_resume_checkpoint_id(resume: Any | None = None) -> str | None:
    checkpoint_id = getattr(resume, "checkpoint_id", None)
    if checkpoint_id is None and isinstance(resume, Mapping):
        checkpoint_id = resume.get("checkpoint_id")
    return str(checkpoint_id).strip() if checkpoint_id is not None else None


def _get_nested_mapping(config: Mapping[str, Any] | None, key: str) -> Mapping[str, Any] | None:
    if not isinstance(config, Mapping):
        return None
    value = config.get(key)
    if isinstance(value, Mapping):
        return value
    return None


def _get_custom_prompt_mapping(config: Mapping[str, Any] | None) -> Mapping[str, Any] | None:
    if not isinstance(config, Mapping):
        return None

    for key in ("custom_prompts", "custom-prompts"):
        value = config.get(key)
        if isinstance(value, Mapping):
            return dict(value)
    return None


def _build_effective_custom_prompts(config: Mapping[str, Any] | None) -> dict[str, Any] | None:
    prompt_defaults = _get_custom_prompt_mapping(config)
    runtime_config = _get_nested_mapping(config, "runtime_config")
    prompt_overrides = _get_custom_prompt_mapping(runtime_config)
    if prompt_defaults is None and prompt_overrides is None:
        return None

    merged_prompts: dict[str, Any] = {}
    if prompt_defaults is not None:
        merged_prompts.update(prompt_defaults)
    if prompt_overrides is not None:
        merged_prompts.update(prompt_overrides)
    return merged_prompts


def _resolve_runtime_config_input(config: Mapping[str, Any] | None) -> Mapping[str, Any] | None:
    if not isinstance(config, Mapping):
        return None

    resolved: dict[str, Any] = dict(config)
    effective_custom_prompts = _build_effective_custom_prompts(config)
    if effective_custom_prompts is not None:
        resolved["custom_prompts"] = effective_custom_prompts
    return resolved


def _build_runtime_request_payload(
    query: str,
    *,
    config: Mapping[str, Any] | None,
    runtime_config: RuntimeConfig,
    hitl_subquestions: bool = False,
    hitl_query_expansion: bool = False,
    resume: Any | None = None,
) -> RuntimeAgentRunRequest:
    custom_prompts_payload = {
        key: value
        for key in _CUSTOM_PROMPT_KEYS
        if (value := getattr(runtime_config.custom_prompts, key)) is not None
    }
    controls = (
        RuntimeAgentRunControls.model_validate(config.get("controls"))
        if isinstance(config, Mapping) and isinstance(config.get("controls"), Mapping)
        else None
    )
    if hitl_subquestions or hitl_query_expansion:
        controls = _apply_public_hitl_input(
            hitl_subquestions=hitl_subquestions,
            hitl_query_expansion=hitl_query_expansion,
            controls=controls,
        )
    return RuntimeAgentRunRequest(
        query=query,
        thread_id=_resolve_request_thread_id(config) or _resolve_resume_checkpoint_id(resume),
        controls=controls,
        custom_prompts=(
            custom_prompts_payload
            if custom_prompts_payload
            else None
        ),
    )


def _hitl_requested(payload: RuntimeAgentRunRequest, *, resume: Any | None = None) -> bool:
    if resume is not None:
        return True
    return bool(payload.controls and payload.controls.hitl and payload.controls.hitl.enabled)


def _apply_public_hitl_input(
    *,
    hitl_subquestions: bool,
    hitl_query_expansion: bool,
    controls: RuntimeAgentRunControls | None,
) -> RuntimeAgentRunControls:
    existing_hitl = controls.hitl if controls is not None else None
    resolved_hitl = RuntimeHitlControl(
        enabled=existing_hitl.enabled if existing_hitl is not None else False,
        subquestions=(
            RuntimeSubquestionHitlControl(enabled=True)
            if hitl_subquestions
            else existing_hitl.subquestions if existing_hitl is not None else None
        ),
        query_expansion=(
            RuntimeQueryExpansionHitlControl(enabled=True)
            if hitl_query_expansion
            else existing_hitl.query_expansion if existing_hitl is not None else None
        ),
    )
    if controls is None:
        return RuntimeAgentRunControls(hitl=resolved_hitl)
    return controls.model_copy(update={"hitl": resolved_hitl})


def _extract_checkpoint_id(payload: Any) -> str | None:
    if not isinstance(payload, Mapping):
        return None
    raw_checkpoint_id = payload.get("checkpoint_id")
    if raw_checkpoint_id is not None:
        return str(raw_checkpoint_id)
    config = payload.get("config")
    if not isinstance(config, Mapping):
        return None
    configurable = config.get("configurable")
    if not isinstance(configurable, Mapping):
        return None
    raw_checkpoint_id = configurable.get("checkpoint_id") or configurable.get("thread_id")
    if raw_checkpoint_id is None:
        return None
    return str(raw_checkpoint_id)


def _extract_interrupt_payload(payload: Any) -> Any | None:
    if not isinstance(payload, Mapping):
        return None
    interrupts = payload.get("interrupts")
    if isinstance(interrupts, list) and interrupts:
        first_interrupt = interrupts[0]
        if isinstance(first_interrupt, Mapping):
            return first_interrupt.get("value")
    interrupt_update = payload.get("__interrupt__")
    if isinstance(interrupt_update, (list, tuple)) and interrupt_update:
        first_interrupt = interrupt_update[0]
        value = getattr(first_interrupt, "value", None)
        if value is not None:
            return value
    return None


def _normalize_hitl_review(payload: Any) -> HitlReview | None:
    if payload is None:
        return None
    return HitlReview.from_interrupt_payload(payload)


def _translate_sdk_resume(resume: Any | None) -> Any | None:
    if not isinstance(resume, HitlResumeRequest):
        return resume
    if resume.review_kind == "subquestion_review":
        return RuntimeSubquestionResumeEnvelope(
            checkpoint_id=resume.checkpoint_id,
            decisions=[
                RuntimeSubquestionDecision(
                    subquestion_id=decision.item_id,
                    action="deny" if decision.action == "reject" else decision.action,
                    edited_text=decision.replacement_text,
                )
                for decision in resume.decisions
            ],
        )
    if resume.review_kind == "query_expansion_review":
        return RuntimeQueryExpansionResumeEnvelope(
            checkpoint_id=resume.checkpoint_id,
            decisions=[
                RuntimeQueryExpansionDecision(
                    expansion_id=decision.item_id,
                    action="deny" if decision.action == "reject" else decision.action,
                    edited_query=decision.replacement_text,
                )
                for decision in resume.decisions
            ],
        )
    raise SDKConfigurationError(f"Unsupported HITL review kind '{resume.review_kind}'.")


class _RuntimeGraphBuilder:
    def __init__(self, *, context: RuntimeGraphContext) -> None:
        self._context = context

    def compile(self, **compile_kwargs: Any) -> Any:
        return build_runtime_graph(context=self._context, **compile_kwargs)


def _run_hitl_runtime_agent(
    payload: RuntimeAgentRunRequest,
    *,
    model: Any,
    vector_store: Any,
    callbacks: list[Any] | None = None,
    langfuse_callback: Any | None = None,
    resume: Any | None = None,
) -> RuntimeAgentRunResult:
    translated_resume = _translate_sdk_resume(resume)
    run_metadata = legacy_service.build_graph_run_metadata(thread_id=payload.thread_id)
    context = RuntimeGraphContext(
        payload=payload,
        model=model,
        vector_store=vector_store,
        callbacks=list(callbacks or []),
        langfuse_callback=langfuse_callback,
        initial_search_context=[],
    )
    builder = _RuntimeGraphBuilder(context=context)
    graph_input: Any = (
        to_runtime_graph_state(payload, run_metadata=run_metadata, initial_search_context=[])
        if translated_resume is None
        else build_resume_command(translated_resume)
    )
    execution_config = {"configurable": {"thread_id": run_metadata.thread_id}}
    terminal_state: Any = None
    latest_checkpoint_id: str | None = None
    interrupt_payload: Any | None = None
    with compile_graph_with_checkpointer(builder) as graph:
        for item in graph.stream(
            graph_input,
            config=execution_config,
            stream_mode=["values", "tasks", "updates", "checkpoints"],
        ):
            if isinstance(item, tuple) and len(item) == 2:
                mode, payload_item = item
            else:
                mode, payload_item = "values", item
            latest_checkpoint_id = _extract_checkpoint_id(payload_item) or latest_checkpoint_id
            interrupt_payload = _extract_interrupt_payload(payload_item) or interrupt_payload
            if mode == "values":
                terminal_state = payload_item
        if interrupt_payload is not None:
            return RuntimeAgentRunResult(
                status="paused",
                thread_id=run_metadata.thread_id,
                checkpoint_id=latest_checkpoint_id or run_metadata.thread_id,
                review=_normalize_hitl_review(interrupt_payload),
            )
        if terminal_state is None:
            terminal_state = graph.invoke(graph_input, config=execution_config)
    response = legacy_service.map_graph_state_to_runtime_response(terminal_state)
    return RuntimeAgentRunResult(
        status="completed",
        thread_id=run_metadata.thread_id,
        checkpoint_id=latest_checkpoint_id or run_metadata.thread_id,
        response=response,
    )


def _map_sdk_error(*, operation: str, exc: Exception) -> SDKError:
    if isinstance(exc, SDKError):
        return exc
    if isinstance(exc, ValidationError):
        return SDKConfigurationError(f"{operation} failed due to invalid SDK input or configuration.")
    if isinstance(exc, (TimeoutError, FuturesTimeoutError)):
        return SDKTimeoutError(f"{operation} timed out.")

    message = str(exc).lower()
    if any(token in message for token in ("vector", "retriev", "document search", "similarity search")):
        return SDKRetrievalError(f"{operation} failed during retrieval.")
    if any(token in message for token in ("model", "llm", "openai", "completion", "chat")):
        return SDKModelError(f"{operation} failed during model execution.")
    if any(token in message for token in ("config", "invalid", "missing", "required", "argument")):
        return SDKConfigurationError(f"{operation} failed due to invalid SDK input or configuration.")

    return SDKError(f"{operation} failed.")


def advanced_rag(
    query: str,
    *,
    vector_store: Any,
    model: Any,
    hitl_subquestions: bool = False,
    hitl_query_expansion: bool = False,
    config: dict[str, Any] | None = None,
    callbacks: list[Any] | None = None,
    langfuse_callback: Any | None = None,
    langfuse_settings: Mapping[str, Any] | None = None,
    resume: Any | None = None,
) -> RuntimeAgentRunResponse | RuntimeAgentRunResult:
    logger.info(
        "SDK advanced_rag requested query_len=%s vector_store_type=%s model_type=%s hitl_subquestions=%s hitl_query_expansion=%s has_config=%s has_callbacks=%s has_langfuse_callback=%s has_langfuse_settings=%s",
        len(query),
        type(vector_store).__name__,
        type(model).__name__,
        hitl_subquestions,
        hitl_query_expansion,
        config is not None,
        bool(callbacks),
        langfuse_callback is not None,
        langfuse_settings is not None,
    )
    runtime_config = RuntimeConfig.from_dict(_resolve_runtime_config_input(config))
    logger.info(
        "SDK sync runtime config resolved initial_k=%s rerank_enabled=%s rerank_provider=%s",
        runtime_config.retrieval.initial_search_context_k,
        runtime_config.rerank.enabled,
        runtime_config.rerank.provider,
    )
    if model is None:
        logger.error("SDK sync run rejected missing model")
        raise SDKConfigurationError("model is required and cannot be None")
    if vector_store is None:
        logger.error("SDK sync run rejected missing vector_store")
        raise SDKConfigurationError("vector_store is required and cannot be None")
    compatible_vector_store = assert_vector_store_compatible(vector_store)
    logger.info(
        "SDK sync run vector_store validated vector_store_type=%s",
        type(compatible_vector_store).__name__,
    )
    if langfuse_settings is not None and langfuse_callback is None:
        logger.warning(
            "SDK advanced_rag received langfuse_settings without langfuse_callback; "
            "settings are ignored and tracing remains disabled"
        )

    try:
        resolved_callbacks = list(callbacks or [])
        if langfuse_callback is not None:
            resolved_callbacks.append(langfuse_callback)
        request_payload = _build_runtime_request_payload(
            query,
            config=config,
            runtime_config=runtime_config,
            hitl_subquestions=hitl_subquestions,
            hitl_query_expansion=hitl_query_expansion,
            resume=resume,
        )
        if _hitl_requested(request_payload, resume=resume):
            response = _run_hitl_runtime_agent(
                request_payload,
                model=model,
                vector_store=compatible_vector_store,
                callbacks=resolved_callbacks or None,
                langfuse_callback=langfuse_callback,
                resume=resume,
            )
        else:
            response = run_runtime_agent(
                request_payload,
                model=model,
                vector_store=compatible_vector_store,
                callbacks=resolved_callbacks or None,
                langfuse_callback=langfuse_callback,
            )
    except Exception as exc:  # noqa: BLE001
        mapped = _map_sdk_error(operation="advanced_rag", exc=exc)
        logger.exception(
            "SDK advanced_rag failed mapped_error=%s original_error=%s",
            type(mapped).__name__,
            type(exc).__name__,
        )
        raise mapped from exc
    logger.info(
        "SDK advanced_rag completed status=%s sub_qa_count=%s output_len=%s",
        getattr(response, "status", "completed"),
        (
            len(response.response.sub_qa)
            if isinstance(response, RuntimeAgentRunResult) and response.response is not None
            else 0 if isinstance(response, RuntimeAgentRunResult)
            else len(response.sub_qa)
        ),
        (
            len(response.response.output)
            if isinstance(response, RuntimeAgentRunResult) and response.response is not None
            else 0 if isinstance(response, RuntimeAgentRunResult)
            else len(response.output)
        ),
    )
    return response



from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import Any

from langgraph.graph import END, START, StateGraph
from langgraph.types import RetryPolicy, interrupt

from agent_search.runtime.graph.routes import route_post_decompose, route_post_expand, route_subquestion_lanes
from agent_search.runtime.graph.state import (
    HITL_MODE_DISABLED,
    HITL_STAGE_QUERY_EXPANSION,
    HITL_STAGE_SUBQUESTIONS,
    RuntimeGraphContext,
    RuntimeGraphState,
)
from agent_search.runtime.reducers import merge_stage_snapshots
from agent_search.runtime.state import to_rag_state
from schemas import (
    AnswerSubquestionNodeInput,
    DecomposeNodeInput,
    ExpandNodeInput,
    ExpandNodeOutput,
    GraphStageSnapshot,
    RerankNodeInput,
    RuntimeAgentRunRequest,
    SearchNodeInput,
    SubQuestionArtifacts,
    SynthesizeFinalNodeInput,
)
from services import agent_service as legacy_service

_RETRY_POLICY = RetryPolicy(max_attempts=3)


def _attach_checkpoint_metadata(interrupt_payload: Any, *, checkpoint_id: str | None = None) -> Any:
    if not isinstance(interrupt_payload, Mapping):
        return interrupt_payload
    normalized_payload = dict(interrupt_payload)
    if checkpoint_id and not normalized_payload.get("checkpoint_id"):
        normalized_payload["checkpoint_id"] = checkpoint_id
    return normalized_payload


def _apply_subquestion_resume_decisions(
    subquestions: Sequence[str],
    *,
    resume: Any = True,
    interrupt_payload: Any | None = None,
) -> list[str]:
    baseline_subquestions = [str(subquestion) for subquestion in subquestions]
    if not hasattr(resume, "decisions"):
        return baseline_subquestions

    available_ids = [f"sq-{index + 1}" for index in range(len(baseline_subquestions))]
    if isinstance(interrupt_payload, Mapping):
        raw_items = interrupt_payload.get("subquestions")
        if isinstance(raw_items, Sequence) and not isinstance(raw_items, (str, bytes, bytearray)):
            available_ids = []
            for fallback_index, raw_item in enumerate(raw_items):
                if not isinstance(raw_item, Mapping):
                    available_ids.append(f"sq-{fallback_index + 1}")
                    continue
                raw_id = raw_item.get("subquestion_id")
                available_ids.append(str(raw_id or f"sq-{fallback_index + 1}"))

    decisions_by_id = {str(decision.subquestion_id): decision for decision in resume.decisions}
    resolved_subquestions: list[str] = []
    for index, current_text in enumerate(baseline_subquestions):
        decision = decisions_by_id.get(available_ids[index] if index < len(available_ids) else f"sq-{index + 1}")
        if decision is None or decision.action in {"approve", "skip"}:
            resolved_subquestions.append(current_text)
            continue
        if decision.action == "edit":
            edited_text = str(decision.edited_text or "").strip()
            if edited_text:
                resolved_subquestions.append(edited_text)
            continue
        if decision.action == "deny":
            continue
        raise ValueError(f"Unsupported resume action '{decision.action}'.")
    return resolved_subquestions


def _apply_query_expansion_resume_decisions(
    expansions: Sequence[str],
    *,
    resume: Any = True,
    interrupt_payload: Any | None = None,
) -> list[str]:
    baseline_expansions = [str(expansion) for expansion in expansions]
    if not hasattr(resume, "decisions"):
        return baseline_expansions

    available_ids = [f"qe-{index + 1}" for index in range(len(baseline_expansions))]
    if isinstance(interrupt_payload, Mapping):
        raw_items = interrupt_payload.get("expansions")
        if isinstance(raw_items, Sequence) and not isinstance(raw_items, (str, bytes, bytearray)):
            available_ids = []
            for fallback_index, raw_item in enumerate(raw_items):
                if not isinstance(raw_item, Mapping):
                    available_ids.append(f"qe-{fallback_index + 1}")
                    continue
                raw_id = raw_item.get("expansion_id")
                available_ids.append(str(raw_id or f"qe-{fallback_index + 1}"))

    decisions_by_id = {str(decision.expansion_id): decision for decision in resume.decisions}
    resolved_expansions: list[str] = []
    for index, current_query in enumerate(baseline_expansions):
        decision = decisions_by_id.get(available_ids[index] if index < len(available_ids) else f"qe-{index + 1}")
        if decision is None or decision.action in {"approve", "skip"}:
            resolved_expansions.append(current_query)
            continue
        if decision.action == "edit":
            edited_query = str(decision.edited_query or "").strip()
            if edited_query:
                resolved_expansions.append(edited_query)
            continue
        if decision.action == "deny":
            continue
        raise ValueError(f"Unsupported resume action '{decision.action}'.")
    return resolved_expansions


def _append_snapshot(
    state: RuntimeGraphState,
    *,
    stage: str,
    sub_question: str = "",
) -> RuntimeGraphState:
    next_state = to_rag_state(state)
    next_state["stage_snapshots"] = merge_stage_snapshots(
        next_state["stage_snapshots"],
        [
            GraphStageSnapshot(
                stage=stage,
                status="completed",
                sub_question=sub_question,
                decomposition_sub_questions=list(next_state["decomposition_sub_questions"]),
                sub_qa=[item.model_copy(deep=True) for item in next_state["sub_qa"]],
                sub_question_artifacts=[item.model_copy(deep=True) for item in next_state["sub_question_artifacts"]],
                output=next_state["output"],
            )
        ],
    )
    return RuntimeGraphState(
        **next_state,
        initial_search_context=list(state.get("initial_search_context", [])),
        hitl_mode=state.get("hitl_mode", HITL_MODE_DISABLED),
        hitl_checkpoint_stages=tuple(state.get("hitl_checkpoint_stages", ())),
    )


def _with_runtime_context(state: RuntimeGraphState, next_state: dict[str, Any]) -> RuntimeGraphState:
    return RuntimeGraphState(
        **next_state,
        initial_search_context=list(state.get("initial_search_context", [])),
        hitl_mode=state.get("hitl_mode", HITL_MODE_DISABLED),
        hitl_checkpoint_stages=tuple(state.get("hitl_checkpoint_stages", ())),
    )


def _resolve_lane_sub_question(state: RuntimeGraphState) -> str:
    if not state["decomposition_sub_questions"]:
        raise ValueError("Lane state must contain exactly one decomposition sub-question.")
    return state["decomposition_sub_questions"][0]


def _build_decompose_node(context: RuntimeGraphContext):
    def _decompose(state: RuntimeGraphState) -> RuntimeGraphState:
        node_output = legacy_service.run_decomposition_node(
            node_input=DecomposeNodeInput(
                main_question=state["main_question"],
                run_metadata=state["run_metadata"],
                initial_search_context=list(state.get("initial_search_context", [])),
            ),
            model=context.model,
            timeout_s=legacy_service._RUNTIME_TIMEOUT_CONFIG.decomposition_llm_timeout_s,
            callbacks=context.callbacks,
        )
        next_state = legacy_service.apply_decompose_node_output_to_graph_state(state=state, node_output=node_output)
        return _append_snapshot(_with_runtime_context(state, to_rag_state(next_state)), stage="decompose")

    return _decompose


def _build_subquestion_checkpoint_node():
    def _subquestion_checkpoint(state: RuntimeGraphState) -> RuntimeGraphState:
        if HITL_STAGE_SUBQUESTIONS not in state["hitl_checkpoint_stages"] or not state["decomposition_sub_questions"]:
            return state
        interrupt_payload = _attach_checkpoint_metadata(
            {
                "kind": "subquestion_review",
                "stage": "subquestions_ready",
                "subquestions": [
                    {
                        "subquestion_id": f"sq-{index + 1}",
                        "sub_question": sub_question,
                        "index": index,
                    }
                    for index, sub_question in enumerate(state["decomposition_sub_questions"])
                ],
            },
            checkpoint_id=state["run_metadata"].thread_id,
        )
        resume_value = interrupt(interrupt_payload)
        next_state = to_rag_state(state)
        next_state["decomposition_sub_questions"] = _apply_subquestion_resume_decisions(
            state["decomposition_sub_questions"],
            resume=resume_value,
            interrupt_payload=interrupt_payload,
        )
        return _with_runtime_context(state, next_state)

    return _subquestion_checkpoint


def _build_query_expansion_checkpoint_node():
    def _query_expansion_checkpoint(state: RuntimeGraphState) -> RuntimeGraphState:
        if HITL_STAGE_QUERY_EXPANSION not in state["hitl_checkpoint_stages"]:
            return state
        sub_question = _resolve_lane_sub_question(state)
        artifact = next(
            (item for item in state["sub_question_artifacts"] if item.sub_question == sub_question),
            SubQuestionArtifacts(sub_question=sub_question),
        )
        if not artifact.expanded_queries:
            return state
        interrupt_payload = _attach_checkpoint_metadata(
            {
                "kind": "query_expansion_review",
                "stage": "query_expansions_ready",
                "sub_question": sub_question,
                "expansions": [
                    {
                        "expansion_id": f"qe-{index + 1}",
                        "query": expanded_query,
                        "sub_question": sub_question,
                        "index": index,
                    }
                    for index, expanded_query in enumerate(artifact.expanded_queries)
                ],
            },
            checkpoint_id=state["run_metadata"].thread_id,
        )
        resume_value = interrupt(interrupt_payload)
        resolved_queries = _apply_query_expansion_resume_decisions(
            artifact.expanded_queries,
            resume=resume_value,
            interrupt_payload=interrupt_payload,
        )
        next_state = legacy_service.apply_expand_node_output_to_graph_state(
            state=state,
            sub_question=sub_question,
            node_output=ExpandNodeOutput(expanded_queries=resolved_queries),
        )
        return _with_runtime_context(state, to_rag_state(next_state))

    return _query_expansion_checkpoint


def _build_expand_node(context: RuntimeGraphContext):
    def _expand(state: RuntimeGraphState) -> RuntimeGraphState:
        sub_question = _resolve_lane_sub_question(state)
        node_output = legacy_service.run_expand_node(
            node_input=ExpandNodeInput(
                main_question=state["main_question"],
                sub_question=sub_question,
                run_metadata=state["run_metadata"],
            ),
            model=context.model,
            callbacks=context.callbacks,
        )
        next_state = legacy_service.apply_expand_node_output_to_graph_state(
            state=state,
            sub_question=sub_question,
            node_output=node_output,
        )
        return _append_snapshot(_with_runtime_context(state, to_rag_state(next_state)), stage="expand", sub_question=sub_question)

    return _expand


def _build_search_node(context: RuntimeGraphContext):
    def _search(state: RuntimeGraphState) -> RuntimeGraphState:
        sub_question = _resolve_lane_sub_question(state)
        artifact = next(
            (item for item in state["sub_question_artifacts"] if item.sub_question == sub_question),
            SubQuestionArtifacts(sub_question=sub_question),
        )
        node_output = legacy_service.run_search_node(
            node_input=SearchNodeInput(
                sub_question=sub_question,
                expanded_queries=list(artifact.expanded_queries),
                run_metadata=state["run_metadata"],
            ),
            vector_store=context.vector_store,
            k_fetch=legacy_service._SEARCH_NODE_K_FETCH,
        )
        next_state = legacy_service.apply_search_node_output_to_graph_state(
            state=state,
            sub_question=sub_question,
            node_output=node_output,
        )
        return _append_snapshot(_with_runtime_context(state, to_rag_state(next_state)), stage="search", sub_question=sub_question)

    return _search


def _build_rerank_node(context: RuntimeGraphContext):
    def _rerank(state: RuntimeGraphState) -> RuntimeGraphState:
        sub_question = _resolve_lane_sub_question(state)
        artifact = next(
            (item for item in state["sub_question_artifacts"] if item.sub_question == sub_question),
            SubQuestionArtifacts(sub_question=sub_question),
        )
        node_output = legacy_service.run_rerank_node(
            node_input=RerankNodeInput(
                sub_question=sub_question,
                expanded_query=legacy_service._select_compat_expanded_query(
                    sub_question=sub_question,
                    expanded_queries=list(artifact.expanded_queries),
                ),
                retrieved_docs=[row.model_copy(deep=True) for row in artifact.retrieved_docs],
                run_metadata=state["run_metadata"],
            ),
            config=legacy_service._RERANKER_CONFIG,
            callbacks=context.callbacks,
        )
        next_state = legacy_service.apply_rerank_node_output_to_graph_state(
            state=state,
            sub_question=sub_question,
            node_output=node_output,
        )
        return _append_snapshot(_with_runtime_context(state, to_rag_state(next_state)), stage="rerank", sub_question=sub_question)

    return _rerank


def _build_answer_node(context: RuntimeGraphContext):
    def _answer(state: RuntimeGraphState) -> RuntimeGraphState:
        sub_question = _resolve_lane_sub_question(state)
        artifact = next(
            (item for item in state["sub_question_artifacts"] if item.sub_question == sub_question),
            SubQuestionArtifacts(sub_question=sub_question),
        )
        answer_rows = artifact.reranked_docs or artifact.retrieved_docs
        answer_citation_rows = artifact.citation_rows_by_index
        node_output = legacy_service.run_answer_subquestion_node(
            node_input=AnswerSubquestionNodeInput(
                sub_question=sub_question,
                reranked_docs=[row.model_copy(deep=True) for row in answer_rows],
                citation_rows_by_index={
                    key: value.model_copy(deep=True) for key, value in answer_citation_rows.items()
                },
                run_metadata=state["run_metadata"],
            ),
            callbacks=context.callbacks,
        )
        next_state = legacy_service.apply_answer_subquestion_node_output_to_graph_state(
            state=state,
            sub_question=sub_question,
            node_output=node_output,
        )
        return _append_snapshot(_with_runtime_context(state, to_rag_state(next_state)), stage="answer", sub_question=sub_question)

    return _answer


def _build_synthesize_node(context: RuntimeGraphContext):
    def _synthesize(state: RuntimeGraphState) -> RuntimeGraphState:
        node_output = legacy_service.run_synthesize_final_node(
            node_input=SynthesizeFinalNodeInput(
                main_question=state["main_question"],
                sub_qa=[item.model_copy(deep=True) for item in state["sub_qa"]],
                sub_question_artifacts=[item.model_copy(deep=True) for item in state["sub_question_artifacts"]],
                run_metadata=state["run_metadata"],
            ),
            callbacks=context.callbacks,
        )
        next_state = legacy_service.apply_synthesize_final_node_output_to_graph_state(
            state=state,
            node_output=node_output,
        )
        return _append_snapshot(_with_runtime_context(state, to_rag_state(next_state)), stage="synthesize_final")

    return _synthesize


def build_runtime_graph(
    *,
    context: RuntimeGraphContext | None = None,
    **compile_kwargs: Any,
) -> Any:
    resolved_context = context or RuntimeGraphContext(payload=RuntimeAgentRunRequest(query="bootstrap"))
    builder = StateGraph(RuntimeGraphState)
    builder.add_node("decompose", _build_decompose_node(resolved_context))
    builder.add_node("subquestion_checkpoint", _build_subquestion_checkpoint_node())
    builder.add_node("query_expansion_checkpoint", _build_query_expansion_checkpoint_node())
    builder.add_node("expand", _build_expand_node(resolved_context))
    builder.add_node("search", _build_search_node(resolved_context), retry_policy=_RETRY_POLICY)
    builder.add_node("rerank", _build_rerank_node(resolved_context), retry_policy=_RETRY_POLICY)
    builder.add_node("answer", _build_answer_node(resolved_context), retry_policy=_RETRY_POLICY)
    builder.add_node("synthesize", _build_synthesize_node(resolved_context), retry_policy=_RETRY_POLICY)
    builder.add_edge(START, "decompose")
    builder.add_conditional_edges(
        "decompose",
        route_post_decompose,
        {"synthesize": "synthesize", "subquestion_checkpoint": "subquestion_checkpoint"},
    )
    builder.add_conditional_edges("subquestion_checkpoint", route_subquestion_lanes)
    builder.add_conditional_edges(
        "expand",
        route_post_expand,
        {"query_expansion_checkpoint": "query_expansion_checkpoint", "search": "search"},
    )
    builder.add_edge("query_expansion_checkpoint", "search")
    builder.add_edge("search", "rerank")
    builder.add_edge("rerank", "answer")
    builder.add_edge("answer", "synthesize")
    builder.add_edge("synthesize", END)
    return builder.compile(**compile_kwargs)


__all__ = ["build_runtime_graph"]



from __future__ import annotations

import logging
import os
import re
from typing import Any

from langchain_openai import ChatOpenAI

from services.document_validation_service import RetrievedDocument, parse_retrieved_documents

logger = logging.getLogger(__name__)

_SUBANSWER_MODEL = os.getenv("SUBANSWER_MODEL", "gpt-4.1-mini")
_SUBANSWER_TEMPERATURE = float(os.getenv("SUBANSWER_TEMPERATURE", "0"))
_OPENAI_API_KEY = os.getenv("OPENAI_API_KEY", "").strip()
_CITATION_PATTERN = re.compile(r"\[\d+\]")


def _build_context_block(documents: list[RetrievedDocument]) -> str:
    lines: list[str] = []
    for doc in documents:
        lines.append(
            f"[{doc.rank}] title={doc.title} source={doc.source} content={doc.content}"
        )
    return "\n".join(lines)


def _build_fallback_subanswer(*, sub_question: str, documents: list[RetrievedDocument]) -> str:
    if not documents:
        return "No relevant evidence found in reranked documents."
    top_doc = documents[0]
    return f"{top_doc.content} (source: {top_doc.source})"


def generate_subanswer(
    *,
    sub_question: str,
    reranked_retrieved_output: str,
    callbacks: list[Any] | None = None,
) -> str:
    documents = parse_retrieved_documents(reranked_retrieved_output)
    logger.info(
        "Subanswer generation parsed reranked docs sub_question=%s doc_count=%s",
        sub_question,
        len(documents),
    )
    if not documents:
        logger.info(
            "Subanswer generation skipped; no parseable reranked docs sub_question=%s",
            sub_question,
        )
        return _build_fallback_subanswer(sub_question=sub_question, documents=documents)

    context_block = _build_context_block(documents)
    fallback_answer = _build_fallback_subanswer(sub_question=sub_question, documents=documents)
    logger.info(
        "Subanswer generation context prepared sub_question=%s context_lines=%s",
        sub_question,
        len(context_block.splitlines()) if context_block else 0,
    )

    if not _OPENAI_API_KEY:
        logger.info(
            "Subanswer generation using fallback; OPENAI_API_KEY is not set sub_question=%s",
            sub_question,
        )
        return fallback_answer

    try:
        llm = ChatOpenAI(model=_SUBANSWER_MODEL, temperature=_SUBANSWER_TEMPERATURE)
        prompt = (
            "You answer one sub-question using the full reranked evidence list below.\n"
            "Requirements:\n"
            "- Use only the evidence provided below.\n"
            "- Treat each evidence line index as a citation key and cite claims with [index], e.g. [1] or [2][3].\n"
            "- Keep citation indices from the provided evidence lines; do not invent new indices.\n"
            "- Keep it to 1-3 sentences.\n"
            "- Do not summarize the evidence list; directly answer the sub-question using cited evidence.\n"
            "- If evidence is insufficient, explicitly say so.\n\n"
            f"Sub-question:\n{sub_question}\n\n"
            f"Reranked evidence:\n{context_block}\n"
        )
        invoke_config = {"callbacks": callbacks} if callbacks else None
        response = llm.invoke(prompt, config=invoke_config) if invoke_config else llm.invoke(prompt)
        answer = (response.content or "").strip() if hasattr(response, "content") else ""
        if answer:
            logger.info(
                "Subanswer generation LLM success sub_question=%s answer_len=%s citation_refs=%s",
                sub_question,
                len(answer),
                len(_CITATION_PATTERN.findall(answer)),
            )
            return answer
        logger.warning(
            "Subanswer generation returned empty LLM response; using fallback sub_question=%s",
            sub_question,
        )
    except Exception:
        logger.exception(
            "Subanswer generation LLM call failed; using fallback sub_question=%s model=%s",
            sub_question,
            _SUBANSWER_MODEL,
        )

    return fallback_answer



from __future__ import annotations

import logging
import os
import re
from typing import Any
from dataclasses import dataclass

from langchain_classic.retrievers.multi_query import MultiQueryRetriever
from langchain_core.callbacks.manager import CallbackManagerForRetrieverRun
from langchain_core.documents import Document
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.retrievers import BaseRetriever
from langchain_openai import ChatOpenAI

logger = logging.getLogger(__name__)

_RUNTIME_AGENT_MODEL = os.getenv("RUNTIME_AGENT_MODEL", "gpt-4.1-mini")
_OPENAI_API_KEY = os.getenv("OPENAI_API_KEY", "").strip()


@dataclass(frozen=True)
class QueryExpansionConfig:
    model: str
    temperature: float
    max_queries: int
    max_query_length: int


def _read_positive_int(env_key: str, default: int, *, min_value: int = 1) -> int:
    raw_value = os.getenv(env_key, "").strip()
    if not raw_value:
        return default
    try:
        parsed = int(raw_value)
    except ValueError:
        logger.warning(
            "Invalid query expansion config; using default env_key=%s value=%s default=%s",
            env_key,
            raw_value,
            default,
        )
        return default
    if parsed < min_value:
        logger.warning(
            "Out-of-range query expansion config; using default env_key=%s value=%s default=%s",
            env_key,
            parsed,
            default,
        )
        return default
    return parsed


def _read_float(env_key: str, default: float) -> float:
    raw_value = os.getenv(env_key, "").strip()
    if not raw_value:
        return default
    try:
        return float(raw_value)
    except ValueError:
        logger.warning(
            "Invalid query expansion float config; using default env_key=%s value=%s default=%s",
            env_key,
            raw_value,
            default,
        )
        return default


def build_query_expansion_config_from_env() -> QueryExpansionConfig:
    return QueryExpansionConfig(
        model=os.getenv("QUERY_EXPANSION_MODEL", _RUNTIME_AGENT_MODEL).strip() or _RUNTIME_AGENT_MODEL,
        temperature=_read_float("QUERY_EXPANSION_TEMPERATURE", 0.0),
        max_queries=_read_positive_int("QUERY_EXPANSION_MAX_QUERIES", 4, min_value=1),
        max_query_length=_read_positive_int("QUERY_EXPANSION_MAX_QUERY_LENGTH", 256, min_value=16),
    )


def _normalize_query(text: str, *, max_query_length: int) -> str:
    normalized = re.sub(r"\s+", " ", (text or "").strip())
    if not normalized:
        return ""
    if len(normalized) > max_query_length:
        return normalized[:max_query_length].strip()
    return normalized


def _normalize_query_candidates(
    *,
    original_query: str,
    generated_queries: list[str],
    config: QueryExpansionConfig,
) -> list[str]:
    normalized_queries: list[str] = []
    seen: set[str] = set()
    candidates = [original_query, *generated_queries]
    for candidate in candidates:
        normalized = _normalize_query(candidate, max_query_length=config.max_query_length)
        if not normalized:
            continue
        key = normalized.casefold()
        if key in seen:
            continue
        normalized_queries.append(normalized)
        seen.add(key)
        if len(normalized_queries) >= config.max_queries:
            break

    if normalized_queries:
        return normalized_queries
    fallback_query = _normalize_query(original_query, max_query_length=config.max_query_length)
    return [fallback_query] if fallback_query else []


class _NoopRetriever(BaseRetriever):
    def _get_relevant_documents(self, query: str, *, run_manager) -> list[Document]:
        _ = query, run_manager
        return []

    async def _aget_relevant_documents(self, query: str, *, run_manager) -> list[Document]:
        _ = query, run_manager
        return []


def expand_queries_for_subquestion(
    *,
    sub_question: str,
    model: BaseChatModel | None = None,
    config: QueryExpansionConfig | None = None,
    callbacks: list[Any] | None = None,
) -> list[str]:
    effective_config = config or build_query_expansion_config_from_env()
    normalized_sub_question = _normalize_query(sub_question, max_query_length=effective_config.max_query_length)
    if not normalized_sub_question:
        logger.info("Query expansion skipped; empty sub-question")
        return []

    if model is None and not _OPENAI_API_KEY:
        logger.info(
            "Query expansion fallback to original; OPENAI_API_KEY is not set sub_question=%s",
            normalized_sub_question,
        )
        return [normalized_sub_question]

    try:
        llm = model or ChatOpenAI(
            model=effective_config.model,
            temperature=effective_config.temperature,
        )
        if callbacks:
            with_config = getattr(llm, "with_config", None)
            if callable(with_config):
                llm = llm.with_config({"callbacks": callbacks})
        multi_query_retriever = MultiQueryRetriever.from_llm(
            retriever=_NoopRetriever(),
            llm=llm,
            include_original=False,
        )
        generated = multi_query_retriever.generate_queries(
            normalized_sub_question,
            run_manager=CallbackManagerForRetrieverRun.get_noop_manager(),
        )
        normalized_queries = _normalize_query_candidates(
            original_query=normalized_sub_question,
            generated_queries=generated if isinstance(generated, list) else [],
            config=effective_config,
        )
        logger.info(
            "Query expansion generated sub_question=%s generated_count=%s normalized_count=%s max_queries=%s",
            normalized_sub_question,
            len(generated) if isinstance(generated, list) else 0,
            len(normalized_queries),
            effective_config.max_queries,
        )
        return normalized_queries
    except Exception:
        logger.exception(
            "Query expansion generation failed; using fallback sub_question=%s model=%s",
            normalized_sub_question,
            effective_config.model,
        )
        return [normalized_sub_question]



from __future__ import annotations

import os
import re
import logging
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass

_RETRIEVED_LINE_PATTERN = re.compile(r"^\d+\.\s+title=(.*?)\s+source=(.*?)\s+content=(.*)$")
_WORD_PATTERN = re.compile(r"[a-z0-9]+")
_YEAR_PATTERN = re.compile(r"\b(?:19|20)\d{2}\b")
logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class RetrievedDocument:
    rank: int
    title: str
    source: str
    content: str


@dataclass(frozen=True)
class DocumentValidationConfig:
    min_relevance_score: float = 0.0
    source_allowlist: tuple[str, ...] = ()
    min_year: int | None = None
    max_year: int | None = None
    require_year_when_range_set: bool = False
    max_workers: int = 8


@dataclass(frozen=True)
class ValidatedDocumentResult:
    document: RetrievedDocument
    relevance_score: float
    passed: bool
    rejection_reasons: tuple[str, ...]


@dataclass(frozen=True)
class SubQuestionValidationResult:
    total_documents: int
    valid_documents: list[RetrievedDocument]
    validation_results: list[ValidatedDocumentResult]


def build_document_validation_config_from_env() -> DocumentValidationConfig:
    source_allowlist_raw = os.getenv("DOCUMENT_VALIDATION_SOURCE_ALLOWLIST", "")
    source_allowlist = tuple(
        item.strip().lower()
        for item in source_allowlist_raw.split(",")
        if item.strip()
    )

    min_year_raw = os.getenv("DOCUMENT_VALIDATION_MIN_YEAR")
    max_year_raw = os.getenv("DOCUMENT_VALIDATION_MAX_YEAR")
    max_workers_raw = os.getenv("DOCUMENT_VALIDATION_MAX_WORKERS", "8")
    require_year_raw = os.getenv("DOCUMENT_VALIDATION_REQUIRE_YEAR_WHEN_RANGE_SET", "false")

    return DocumentValidationConfig(
        min_relevance_score=float(os.getenv("DOCUMENT_VALIDATION_MIN_RELEVANCE_SCORE", "0.0")),
        source_allowlist=source_allowlist,
        min_year=int(min_year_raw) if min_year_raw not in (None, "") else None,
        max_year=int(max_year_raw) if max_year_raw not in (None, "") else None,
        require_year_when_range_set=require_year_raw.strip().lower() in {"1", "true", "yes", "on"},
        max_workers=max(1, int(max_workers_raw)),
    )


def parse_retrieved_documents(retrieved_output: str) -> list[RetrievedDocument]:
    if not isinstance(retrieved_output, str) or not retrieved_output.strip():
        return []

    parsed: list[RetrievedDocument] = []
    for line in retrieved_output.splitlines():
        match = _RETRIEVED_LINE_PATTERN.match(line.strip())
        if not match:
            continue
        title, source, content = match.groups()
        parsed.append(
            RetrievedDocument(
                rank=len(parsed) + 1,
                title=title.strip(),
                source=source.strip(),
                content=content.strip(),
            )
        )
    return parsed


def format_retrieved_documents(documents: list[RetrievedDocument]) -> str:
    """Serialize documents using the stable citation contract.

    Contract shape (one line per document):
    ``{index}. title={title} source={source} content={content}``
    """
    if not documents:
        return "No relevant documents found."
    formatted = "\n".join(
        f"{idx}. title={doc.title} source={doc.source} content={doc.content}"
        for idx, doc in enumerate(documents, start=1)
    )
    logger.info(
        "Document validation formatter emitted citation contract document_count=%s contract=%s",
        len(documents),
        "index.title.source.content",
    )
    return formatted


def _tokenize(value: str) -> set[str]:
    return {token for token in _WORD_PATTERN.findall(value.lower()) if token}


def _compute_relevance_score(sub_question: str, document: RetrievedDocument) -> float:
    query_tokens = _tokenize(sub_question)
    if not query_tokens:
        return 0.0
    doc_tokens = _tokenize(" ".join((document.title, document.content, document.source)))
    if not doc_tokens:
        return 0.0
    overlap = len(query_tokens & doc_tokens)
    return overlap / len(query_tokens)


def _extract_years(document: RetrievedDocument) -> list[int]:
    text = " ".join((document.title, document.content, document.source))
    return [int(match.group(0)) for match in _YEAR_PATTERN.finditer(text)]


def _validate_document(
    *,
    sub_question: str,
    document: RetrievedDocument,
    config: DocumentValidationConfig,
) -> ValidatedDocumentResult:
    rejection_reasons: list[str] = []
    relevance_score = _compute_relevance_score(sub_question, document)

    if relevance_score < config.min_relevance_score:
        rejection_reasons.append(
            f"relevance_below_threshold({relevance_score:.3f}<{config.min_relevance_score:.3f})"
        )

    source_allowlist = config.source_allowlist
    if source_allowlist:
        source_value = document.source.lower()
        if source_value not in source_allowlist:
            rejection_reasons.append("source_not_allowlisted")

    range_configured = config.min_year is not None or config.max_year is not None
    if range_configured:
        years = _extract_years(document)
        if not years:
            if config.require_year_when_range_set:
                rejection_reasons.append("missing_year")
        else:
            in_range = False
            for year in years:
                if config.min_year is not None and year < config.min_year:
                    continue
                if config.max_year is not None and year > config.max_year:
                    continue
                in_range = True
                break
            if not in_range:
                rejection_reasons.append("year_out_of_range")

    return ValidatedDocumentResult(
        document=document,
        relevance_score=relevance_score,
        passed=not rejection_reasons,
        rejection_reasons=tuple(rejection_reasons),
    )


def validate_subquestion_documents(
    *,
    sub_question: str,
    retrieved_output: str,
    config: DocumentValidationConfig,
) -> SubQuestionValidationResult:
    documents = parse_retrieved_documents(retrieved_output)
    if not documents:
        return SubQuestionValidationResult(
            total_documents=0,
            valid_documents=[],
            validation_results=[],
        )

    with ThreadPoolExecutor(max_workers=min(config.max_workers, len(documents))) as executor:
        validation_results = list(
            executor.map(
                lambda doc: _validate_document(sub_question=sub_question, document=doc, config=config),
                documents,
            )
        )

    valid_documents = [result.document for result in validation_results if result.passed]
    return SubQuestionValidationResult(
        total_documents=len(documents),
        valid_documents=valid_documents,
        validation_results=validation_results,
    )



from .agent import (
    AgentGraphState,
    AgentRunStageMetadata,
    AnswerSubquestionNodeInput,
    AnswerSubquestionNodeOutput,
    CitationSourceRow,
    DecomposeNodeInput,
    DecomposeNodeOutput,
    ExpandNodeInput,
    ExpandNodeOutput,
    GraphRunMetadata,
    GraphStageSnapshot,
    HitlResumeDecision,
    HitlResumeRequest,
    HitlReview,
    HitlReviewItem,
    RerankNodeInput,
    RerankNodeOutput,
    RuntimeAgentInfo,
    RuntimeAgentRunAsyncCancelResponse,
    RuntimeAgentRunAsyncStartResponse,
    RuntimeAgentRunAsyncStatusResponse,
    RuntimeAgentRunControls,
    RuntimeAgentRunRequest,
    RuntimeAgentRunResumeRequest,
    RuntimeAgentRunResponse,
    RuntimeHitlControl,
    RuntimeQueryExpansionDecision,
    RuntimeQueryExpansionHitlControl,
    RuntimeQueryExpansionResumeEnvelope,
    RuntimeQueryExpansionControl,
    RuntimeRerankControl,
    RuntimeSubquestionDecision,
    RuntimeSubquestionHitlControl,
    RuntimeSubquestionResumeEnvelope,
    SearchNodeInput,
    SearchNodeOutput,
    SubQuestionArtifacts,
    SubQuestionAnswer,
    SynthesizeFinalNodeInput,
    SynthesizeFinalNodeOutput,
)
from .internal_data import (
    InternalDataLoadRequest,
    InternalDataLoadResponse,
    InternalDataLoadJobCancelResponse,
    InternalDataLoadJobStartResponse,
    InternalDataLoadJobStatusResponse,
    InternalDataRetrieveRequest,
    InternalDataRetrieveResponse,
    InternalDocumentInput,
    InternalRetrievedChunk,
    WipeResponse,
    WikiLoadInput,
    WikiSourceOption,
    WikiSourcesResponse,
)

__all__ = [
    "RuntimeAgentRunRequest",
    "RuntimeAgentRunResponse",
    "RuntimeAgentRunAsyncStartResponse",
    "RuntimeAgentRunAsyncStatusResponse",
    "RuntimeAgentRunAsyncCancelResponse",
    "RuntimeAgentRunControls",
    "RuntimeAgentRunResumeRequest",
    "RuntimeAgentInfo",
    "RuntimeHitlControl",
    "HitlResumeDecision",
    "HitlResumeRequest",
    "HitlReview",
    "HitlReviewItem",
    "RuntimeQueryExpansionDecision",
    "RuntimeQueryExpansionHitlControl",
    "RuntimeQueryExpansionResumeEnvelope",
    "RuntimeQueryExpansionControl",
    "RuntimeRerankControl",
    "RuntimeSubquestionDecision",
    "RuntimeSubquestionHitlControl",
    "RuntimeSubquestionResumeEnvelope",
    "AgentRunStageMetadata",
    "SubQuestionAnswer",
    "GraphRunMetadata",
    "GraphStageSnapshot",
    "CitationSourceRow",
    "SubQuestionArtifacts",
    "AgentGraphState",
    "DecomposeNodeInput",
    "DecomposeNodeOutput",
    "ExpandNodeInput",
    "ExpandNodeOutput",
    "SearchNodeInput",
    "SearchNodeOutput",
    "RerankNodeInput",
    "RerankNodeOutput",
    "AnswerSubquestionNodeInput",
    "AnswerSubquestionNodeOutput",
    "SynthesizeFinalNodeInput",
    "SynthesizeFinalNodeOutput",
    "InternalDocumentInput",
    "InternalDataLoadRequest",
    "InternalDataLoadResponse",
    "InternalDataLoadJobCancelResponse",
    "InternalDataLoadJobStartResponse",
    "InternalDataLoadJobStatusResponse",
    "InternalDataRetrieveRequest",
    "InternalDataRetrieveResponse",
    "InternalRetrievedChunk",
    "WipeResponse",
    "WikiLoadInput",
    "WikiSourceOption",
    "WikiSourcesResponse",
]



from __future__ import annotations

import os
from dataclasses import dataclass

from schemas import SubQuestionAnswer

_INSUFFICIENT_ANSWER_PATTERNS = (
    "no relevant docs",
    "no relevant evidence",
    "no relevant information",
    "no information available",
    "insufficient evidence",
    "not enough evidence",
    "cannot determine",
    "unable to determine",
    "insufficient information",
    "nothing relevant found",
    "does not contain any information",
    "does not contain information",
)
_MIN_ANSWERABLE_RATIO = float(os.getenv("REFINEMENT_MIN_ANSWERABLE_RATIO", "0.5"))


@dataclass(frozen=True)
class RefinementDecision:
    refinement_needed: bool
    reason: str


def should_refine(
    *,
    question: str,
    initial_answer: str,
    sub_qa: list[SubQuestionAnswer],
) -> RefinementDecision:
    _ = question  # Reserved for later, richer question-to-answer coverage checks.

    normalized_answer = (initial_answer or "").strip()
    if not normalized_answer:
        return RefinementDecision(refinement_needed=True, reason="initial_answer_empty")

    lowered_answer = normalized_answer.lower()
    for pattern in _INSUFFICIENT_ANSWER_PATTERNS:
        if pattern in lowered_answer:
            return RefinementDecision(
                refinement_needed=True,
                reason="initial_answer_reports_insufficient_evidence",
            )

    if not sub_qa:
        return RefinementDecision(refinement_needed=True, reason="no_subquestion_answers")

    answerable_count = sum(1 for item in sub_qa if item.answerable)
    if answerable_count == 0:
        return RefinementDecision(refinement_needed=True, reason="no_answerable_subanswers")

    answerable_ratio = answerable_count / len(sub_qa)
    if answerable_ratio < _MIN_ANSWERABLE_RATIO:
        return RefinementDecision(
            refinement_needed=True,
            reason=f"low_answerable_ratio:{answerable_ratio:.2f}",
        )

    return RefinementDecision(
        refinement_needed=False,
        reason=f"sufficient_answerable_ratio:{answerable_ratio:.2f}",
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Annotated, Any, Literal, Mapping

from typing_extensions import TypedDict

from agent_search.runtime.state import (
    CitationRowsByIndexChannel,
    DecompositionSubQuestionsChannel,
    StageSnapshotsChannel,
    SubQAChannel,
    SubQuestionArtifactsChannel,
    to_rag_state,
)
from schemas import GraphRunMetadata, RuntimeAgentRunRequest

RuntimeHitlMode = Literal["disabled", "strategic", "custom"]
RuntimeHitlCheckpointStage = Literal["subquestions", "query_expansion"]

HITL_MODE_DISABLED: RuntimeHitlMode = "disabled"
HITL_MODE_STRATEGIC: RuntimeHitlMode = "strategic"
HITL_MODE_CUSTOM: RuntimeHitlMode = "custom"

HITL_STAGE_SUBQUESTIONS: RuntimeHitlCheckpointStage = "subquestions"
HITL_STAGE_QUERY_EXPANSION: RuntimeHitlCheckpointStage = "query_expansion"
STRATEGIC_HITL_CHECKPOINT_STAGES: tuple[RuntimeHitlCheckpointStage, ...] = (
    HITL_STAGE_SUBQUESTIONS,
    HITL_STAGE_QUERY_EXPANSION,
)


def _merge_stable_main_question(current: str, update: str) -> str:
    if current and update and current != update:
        raise ValueError(f"main_question must remain stable across graph lanes: {current!r} != {update!r}")
    return update or current


def _merge_stable_optional_text(current: str, update: str) -> str:
    if current and update and current != update:
        raise ValueError(f"text channel must remain stable across graph lanes: {current!r} != {update!r}")
    return update or current


def _merge_stable_hitl_mode(current: RuntimeHitlMode, update: RuntimeHitlMode) -> RuntimeHitlMode:
    if current and update and current != update:
        raise ValueError(f"hitl_mode must remain stable across graph lanes: {current!r} != {update!r}")
    return update or current


def _merge_stable_run_metadata(current: GraphRunMetadata, update: GraphRunMetadata) -> GraphRunMetadata:
    if current.model_dump(mode="json") != update.model_dump(mode="json"):
        raise ValueError("run_metadata must remain stable across graph lanes.")
    return update.model_copy(deep=True)


def _merge_stable_initial_search_context(
    current: list[dict[str, Any]],
    update: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    if current and update and current != update:
        raise ValueError("initial_search_context must remain stable across graph lanes.")
    return [dict(item) for item in (update or current)]


def _merge_stable_hitl_checkpoint_stages(
    current: tuple[RuntimeHitlCheckpointStage, ...],
    update: tuple[RuntimeHitlCheckpointStage, ...],
) -> tuple[RuntimeHitlCheckpointStage, ...]:
    if current and update and current != update:
        raise ValueError("hitl_checkpoint_stages must remain stable across graph lanes.")
    return update or current


def _resolve_hitl_policy(
    payload: RuntimeAgentRunRequest,
) -> tuple[RuntimeHitlMode, tuple[RuntimeHitlCheckpointStage, ...]]:
    hitl = payload.controls.hitl if payload.controls else None
    if hitl is None or not hitl.enabled:
        return HITL_MODE_DISABLED, ()

    explicit_stages: list[RuntimeHitlCheckpointStage] = []
    if hitl.subquestions and hitl.subquestions.enabled:
        explicit_stages.append(HITL_STAGE_SUBQUESTIONS)
    if hitl.query_expansion and hitl.query_expansion.enabled:
        explicit_stages.append(HITL_STAGE_QUERY_EXPANSION)
    if explicit_stages:
        return HITL_MODE_CUSTOM, tuple(explicit_stages)
    return HITL_MODE_STRATEGIC, STRATEGIC_HITL_CHECKPOINT_STAGES


@dataclass(slots=True)
class RuntimeGraphContext:
    payload: RuntimeAgentRunRequest
    model: Any | None = None
    vector_store: Any | None = None
    callbacks: list[Any] = field(default_factory=list)
    langfuse_callback: Any | None = None
    initial_search_context: list[dict[str, Any]] = field(default_factory=list)


def to_runtime_graph_state(
    payload: RuntimeAgentRunRequest,
    *,
    run_metadata: GraphRunMetadata,
    initial_search_context: list[dict[str, Any]] | None = None,
) -> "RuntimeGraphState":
    hitl_mode, hitl_checkpoint_stages = _resolve_hitl_policy(payload)
    base_state = to_rag_state(
        {
            "main_question": payload.query,
            "decomposition_sub_questions": [],
            "sub_question_artifacts": [],
            "final_answer": "",
            "citation_rows_by_index": {},
            "run_metadata": run_metadata,
            "sub_qa": [],
            "output": "",
            "stage_snapshots": [],
        }
    )
    return RuntimeGraphState(
        **base_state,
        initial_search_context=list(initial_search_context or []),
        hitl_mode=hitl_mode,
        hitl_checkpoint_stages=hitl_checkpoint_stages,
    )


class RuntimeGraphState(TypedDict):
    main_question: Annotated[str, _merge_stable_main_question]
    decomposition_sub_questions: DecompositionSubQuestionsChannel
    sub_question_artifacts: SubQuestionArtifactsChannel
    final_answer: Annotated[str, _merge_stable_optional_text]
    citation_rows_by_index: CitationRowsByIndexChannel
    run_metadata: Annotated[GraphRunMetadata, _merge_stable_run_metadata]
    sub_qa: SubQAChannel
    output: Annotated[str, _merge_stable_optional_text]
    stage_snapshots: StageSnapshotsChannel
    initial_search_context: Annotated[list[dict[str, Any]], _merge_stable_initial_search_context]
    hitl_mode: Annotated[RuntimeHitlMode, _merge_stable_hitl_mode]
    hitl_checkpoint_stages: Annotated[
        tuple[RuntimeHitlCheckpointStage, ...], _merge_stable_hitl_checkpoint_stages
    ]


RuntimeGraphStateMapping = Mapping[str, Any]


__all__ = [
    "RuntimeGraphContext",
    "RuntimeHitlCheckpointStage",
    "RuntimeHitlMode",
    "RuntimeGraphState",
    "RuntimeGraphStateMapping",
    "HITL_MODE_CUSTOM",
    "HITL_MODE_DISABLED",
    "HITL_MODE_STRATEGIC",
    "HITL_STAGE_QUERY_EXPANSION",
    "HITL_STAGE_SUBQUESTIONS",
    "STRATEGIC_HITL_CHECKPOINT_STAGES",
    "to_runtime_graph_state",
]



from __future__ import annotations

from collections.abc import Mapping
from datetime import datetime, timezone
from typing import Any, Callable

from pydantic import BaseModel

from schemas import GraphRunMetadata
from utils.langfuse_tracing import build_trace_metadata


class RuntimeLifecycleEvent(BaseModel):
    event_type: str
    event_id: str
    run_id: str
    thread_id: str
    trace_id: str
    stage: str
    status: str
    emitted_at: str
    error: str | None = None


def _utcnow() -> datetime:
    return datetime.now(timezone.utc)


def _normalize_stage_name(stage_name: str | None) -> str:
    normalized = (stage_name or "").strip()
    if not normalized:
        return "runtime"
    if normalized == "synthesize":
        return "synthesize_final"
    return normalized


class LifecycleEventBuilder:
    def __init__(
        self,
        *,
        run_metadata: GraphRunMetadata,
        clock: Callable[[], datetime] = _utcnow,
    ) -> None:
        self._run_metadata = run_metadata
        self._clock = clock
        self._sequence = 0
        self._active_stage = "runtime"

    @property
    def active_stage(self) -> str:
        return self._active_stage

    def emit_run_started(self) -> RuntimeLifecycleEvent:
        return self._emit(event_type="run.started", stage="runtime", status="running")

    def emit_recovery_started(self, *, checkpoint_id: str | None = None) -> RuntimeLifecycleEvent:
        error = None
        if checkpoint_id:
            error = f"resuming from checkpoint {checkpoint_id}"
        return self._emit(event_type="run.recovered", stage="recovery", status="running", error=error)

    def emit_retrying(self, *, stage: str, error: str) -> RuntimeLifecycleEvent:
        normalized_stage = _normalize_stage_name(stage)
        self._active_stage = normalized_stage
        return self._emit(event_type="stage.retrying", stage=normalized_stage, status="retrying", error=error)

    def emit_terminal(self, *, status: str, error: str | None = None) -> RuntimeLifecycleEvent:
        event_type = {
            "success": "run.completed",
            "completed": "run.completed",
            "error": "run.failed",
            "failed": "run.failed",
            "paused": "run.paused",
        }.get(status, "run.completed")
        return self._emit(event_type=event_type, stage=self._active_stage, status=status, error=error)

    def consume_stream_signal(self, mode: str, payload: Any) -> list[RuntimeLifecycleEvent]:
        if not isinstance(payload, Mapping):
            return []
        if mode == "tasks":
            return self._consume_task_signal(payload)
        if mode == "updates":
            return self._consume_update_signal(payload)
        if mode == "checkpoints":
            return self._consume_checkpoint_signal(payload)
        return []

    def _consume_task_signal(self, payload: Mapping[str, Any]) -> list[RuntimeLifecycleEvent]:
        stage = _normalize_stage_name(str(payload.get("name", "")))
        if "input" in payload:
            self._active_stage = stage
            return [self._emit(event_type="stage.started", stage=stage, status="running")]
        if payload.get("error"):
            self._active_stage = stage
            return [self._emit(event_type="stage.failed", stage=stage, status="error", error=str(payload["error"]))]
        interrupts = payload.get("interrupts")
        if isinstance(interrupts, list) and interrupts:
            self._active_stage = stage
            return [self._emit(event_type="stage.interrupted", stage=stage, status="paused")]
        self._active_stage = stage
        return [self._emit(event_type="stage.completed", stage=stage, status="completed")]

    def _consume_update_signal(self, payload: Mapping[str, Any]) -> list[RuntimeLifecycleEvent]:
        events: list[RuntimeLifecycleEvent] = []
        for stage_name in payload:
            stage = _normalize_stage_name(str(stage_name))
            self._active_stage = stage
            events.append(self._emit(event_type="stage.updated", stage=stage, status="running"))
        return events

    def _consume_checkpoint_signal(self, payload: Mapping[str, Any]) -> list[RuntimeLifecycleEvent]:
        metadata = payload.get("metadata")
        if not isinstance(metadata, Mapping):
            return []
        source = str(metadata.get("source", "")).strip()
        if source == "input":
            return []
        config = payload.get("config")
        checkpoint_id = None
        if isinstance(config, Mapping):
            configurable = config.get("configurable")
            if isinstance(configurable, Mapping):
                raw_checkpoint_id = configurable.get("checkpoint_id")
                if raw_checkpoint_id is not None:
                    checkpoint_id = str(raw_checkpoint_id)
        next_nodes = payload.get("next")
        stage = self._active_stage
        if isinstance(next_nodes, list) and next_nodes:
            stage = _normalize_stage_name(str(next_nodes[0]))
        self._active_stage = stage
        return [
            self._emit(
                event_type="checkpoint.created",
                stage=stage,
                status="checkpointed",
                error=f"checkpoint_id={checkpoint_id}" if checkpoint_id else None,
            )
        ]

    def _emit(
        self,
        *,
        event_type: str,
        stage: str,
        status: str,
        error: str | None = None,
    ) -> RuntimeLifecycleEvent:
        self._sequence += 1
        emitted_at = self._clock().isoformat()
        metadata = build_trace_metadata(
            run_metadata=self._run_metadata,
            stage=stage,
            status=status,
        )
        return RuntimeLifecycleEvent(
            event_type=event_type,
            event_id=f"{self._run_metadata.run_id}:{self._sequence:06d}",
            run_id=str(metadata["run_id"]),
            thread_id=str(metadata["thread_id"]),
            trace_id=str(metadata["trace_id"]),
            stage=str(metadata["stage"]),
            status=str(metadata["status"]),
            emitted_at=emitted_at,
            error=error,
        )


__all__ = ["LifecycleEventBuilder", "RuntimeLifecycleEvent"]



from __future__ import annotations

from collections.abc import Sequence

from langgraph.types import Send

from agent_search.runtime.graph.state import (
    HITL_STAGE_QUERY_EXPANSION,
    HITL_STAGE_SUBQUESTIONS,
    RuntimeGraphState,
    RuntimeHitlCheckpointStage,
)


def _hitl_stage_enabled(state: RuntimeGraphState, stage: RuntimeHitlCheckpointStage) -> bool:
    return stage in state.get("hitl_checkpoint_stages", ())


def route_post_decompose(state: RuntimeGraphState) -> str | Sequence[Send]:
    if not state["decomposition_sub_questions"]:
        return "synthesize"
    if _hitl_stage_enabled(state, HITL_STAGE_SUBQUESTIONS):
        return "subquestion_checkpoint"
    return route_subquestion_lanes(state)


def route_post_expand(state: RuntimeGraphState) -> str:
    if _hitl_stage_enabled(state, HITL_STAGE_QUERY_EXPANSION):
        return "query_expansion_checkpoint"
    return "search"


def route_subquestion_lanes(state: RuntimeGraphState) -> Sequence[Send]:
    return [
        Send(
            "expand",
            {
                "main_question": state["main_question"],
                "decomposition_sub_questions": [sub_question],
                "sub_question_artifacts": [],
                "final_answer": "",
                "citation_rows_by_index": {},
                "run_metadata": state["run_metadata"],
                "sub_qa": [],
                "output": "",
                "stage_snapshots": [],
                "hitl_mode": state["hitl_mode"],
                "hitl_checkpoint_stages": state["hitl_checkpoint_stages"],
            },
        )
        for sub_question in state["decomposition_sub_questions"]
    ]


__all__ = ["route_post_decompose", "route_post_expand", "route_subquestion_lanes"]



from typing import Literal, Optional

from pydantic import BaseModel, Field, model_validator


class InternalDocumentInput(BaseModel):
    title: str = Field(min_length=1)
    content: str = Field(min_length=1)
    source_ref: Optional[str] = None
    source_url: Optional[str] = None  # e.g. wiki page URL for chunk metadata


class WikiLoadInput(BaseModel):
    source_id: str = Field(min_length=1)


class InternalDataLoadRequest(BaseModel):
    source_type: Literal["inline", "wiki"] = "inline"
    documents: Optional[list[InternalDocumentInput]] = None
    wiki: Optional[WikiLoadInput] = None

    @model_validator(mode="after")
    def validate_source_payload(self) -> "InternalDataLoadRequest":
        if self.source_type == "inline":
            if not self.documents:
                raise ValueError("inline load requires at least one document")
            return self

        if self.source_type == "wiki":
            if self.documents:
                raise ValueError("wiki load does not accept inline documents")
            if self.wiki is None:
                raise ValueError("wiki load requires wiki payload")
            return self

        raise ValueError("unsupported source_type")


class InternalDataLoadResponse(BaseModel):
    status: Literal["success"]
    source_type: str
    documents_loaded: int
    chunks_created: int


class InternalDataLoadJobStartResponse(BaseModel):
    job_id: str
    status: str


class InternalDataLoadJobStatusResponse(BaseModel):
    job_id: str
    status: str
    total: int
    completed: int
    message: str
    error: Optional[str] = None
    response: Optional[InternalDataLoadResponse] = None


class InternalDataLoadJobCancelResponse(BaseModel):
    status: Literal["success"]
    message: str


class WipeResponse(BaseModel):
    status: Literal["success"]
    message: str


class InternalDataRetrieveRequest(BaseModel):
    query: str = Field(min_length=1)
    limit: int = Field(default=5, ge=1, le=20)


class InternalRetrievedChunk(BaseModel):
    chunk_id: int
    document_id: int
    document_title: str
    source_type: str
    source_ref: Optional[str]
    content: str
    score: float
    chunk_metadata: Optional[dict[str, str]] = None


class InternalDataRetrieveResponse(BaseModel):
    query: str
    total_chunks_considered: int
    results: list[InternalRetrievedChunk]


class WikiSourceOption(BaseModel):
    source_id: str
    label: str
    article_query: str
    already_loaded: bool


class WikiSourcesResponse(BaseModel):
    sources: list[WikiSourceOption]



from __future__ import annotations

import uuid
from collections.abc import Mapping, Sequence
from typing import Any, Literal, Union

from pydantic import AliasChoices
from pydantic import BaseModel, ConfigDict, Field
from pydantic import field_validator
from pydantic import model_validator


class RuntimeRerankControl(BaseModel):
    enabled: bool | None = None


class RuntimeQueryExpansionControl(BaseModel):
    enabled: bool | None = None


class RuntimeSubquestionHitlControl(BaseModel):
    enabled: bool = False


class RuntimeQueryExpansionHitlControl(BaseModel):
    enabled: bool = False


class RuntimeHitlControl(BaseModel):
    enabled: bool = False
    subquestions: RuntimeSubquestionHitlControl | None = None
    query_expansion: RuntimeQueryExpansionHitlControl | None = None

    @model_validator(mode="after")
    def normalize_enabled(self) -> "RuntimeHitlControl":
        if (
            (self.subquestions is not None and self.subquestions.enabled)
            or (self.query_expansion is not None and self.query_expansion.enabled)
        ):
            self.enabled = True
        elif self.enabled and self.subquestions is None and self.query_expansion is None:
            self.subquestions = RuntimeSubquestionHitlControl(enabled=True)
        return self


class RuntimeAgentRunControls(BaseModel):
    rerank: RuntimeRerankControl | None = None
    query_expansion: RuntimeQueryExpansionControl | None = None
    hitl: RuntimeHitlControl | None = None


class RuntimeAgentRunRuntimeConfig(BaseModel):
    rerank: RuntimeRerankControl | None = None
    query_expansion: RuntimeQueryExpansionControl | None = None


class RuntimeCustomPrompts(BaseModel):
    model_config = ConfigDict(extra="ignore")

    subanswer: str | None = None
    synthesis: str | None = None


class RuntimeAgentRunRequest(BaseModel):
    query: str = Field(min_length=1)
    thread_id: str | None = None
    controls: RuntimeAgentRunControls | None = None
    runtime_config: RuntimeAgentRunRuntimeConfig | None = None
    custom_prompts: RuntimeCustomPrompts | None = Field(
        default=None,
        validation_alias=AliasChoices("custom_prompts", "custom-prompts"),
    )

    @field_validator("thread_id")
    @classmethod
    def validate_thread_id(cls, value: str | None) -> str | None:
        if value is None:
            return None
        normalized_value = value.strip()
        if not normalized_value:
            raise ValueError("thread_id must be a non-empty UUID string.")
        try:
            return str(uuid.UUID(normalized_value))
        except ValueError as exc:
            raise ValueError("thread_id must be a valid UUID string.") from exc


class SubQuestionAnswer(BaseModel):
    sub_question: str
    sub_answer: str
    tool_call_input: str = ""
    expanded_query: str = ""
    sub_agent_response: str = ""
    answerable: bool = False
    verification_reason: str = ""


class RuntimeAgentRunResponse(BaseModel):
    main_question: str = ""
    thread_id: str = ""
    sub_qa: list[SubQuestionAnswer] = Field(default_factory=list)
    sub_answers: list[SubQuestionAnswer] = Field(default_factory=list)
    output: str
    final_citations: list["CitationSourceRow"] = Field(default_factory=list)


class HitlResumeDecision(BaseModel):
    item_id: str = Field(min_length=1)
    action: Literal["approve", "edit", "reject"]
    replacement_text: str | None = None

    @model_validator(mode="after")
    def validate_replacement_text(self) -> "HitlResumeDecision":
        if self.action == "edit":
            if self.replacement_text is None or not self.replacement_text.strip():
                raise ValueError("replacement_text is required when action='edit'.")
            self.replacement_text = self.replacement_text.strip()
            return self
        if self.replacement_text is not None:
            self.replacement_text = self.replacement_text.strip() or None
        return self


class HitlReviewItem(BaseModel):
    item_id: str = Field(min_length=1)
    text: str
    index: int = Field(ge=0)

    def approve(self) -> HitlResumeDecision:
        return HitlResumeDecision(item_id=self.item_id, action="approve")

    def edit(self, replacement_text: str) -> HitlResumeDecision:
        return HitlResumeDecision(item_id=self.item_id, action="edit", replacement_text=replacement_text)

    def reject(self) -> HitlResumeDecision:
        return HitlResumeDecision(item_id=self.item_id, action="reject")


class HitlReview(BaseModel):
    kind: Literal["subquestion_review", "query_expansion_review"]
    stage: str = ""
    checkpoint_id: str = Field(min_length=1)
    items: list[HitlReviewItem] = Field(default_factory=list)

    @classmethod
    def from_interrupt_payload(cls, payload: Any) -> "HitlReview":
        if not isinstance(payload, Mapping):
            raise ValueError("HITL interrupt payload must be a mapping.")
        kind = str(payload.get("kind") or "").strip()
        checkpoint_id = str(payload.get("checkpoint_id") or "").strip()
        if not kind:
            raise ValueError("HITL interrupt payload is missing kind.")
        if kind not in {"subquestion_review", "query_expansion_review"}:
            raise ValueError(f"Unsupported HITL review kind '{kind}'.")
        if not checkpoint_id:
            raise ValueError("HITL interrupt payload is missing checkpoint_id.")
        return cls(
            kind=kind,
            stage=str(payload.get("stage") or ""),
            checkpoint_id=checkpoint_id,
            items=cls._normalize_items(kind=kind, payload=payload),
        )

    @staticmethod
    def _normalize_items(*, kind: str, payload: Mapping[str, Any]) -> list[HitlReviewItem]:
        raw_items_key = "subquestions" if kind == "subquestion_review" else "expansions"
        raw_items = payload.get(raw_items_key)
        if not isinstance(raw_items, Sequence) or isinstance(raw_items, (str, bytes, bytearray)):
            return []
        item_id_key = "subquestion_id" if kind == "subquestion_review" else "expansion_id"
        text_key = "sub_question" if kind == "subquestion_review" else "query"
        items: list[HitlReviewItem] = []
        for index, raw_item in enumerate(raw_items):
            if not isinstance(raw_item, Mapping):
                continue
            raw_index = raw_item.get("index")
            items.append(
                HitlReviewItem(
                    item_id=str(raw_item.get(item_id_key) or f"item-{index + 1}").strip(),
                    text=str(raw_item.get(text_key) or "").strip(),
                    index=raw_index if isinstance(raw_index, int) and raw_index >= 0 else index,
                )
            )
        return items

    def approve_all(self) -> "HitlResumeRequest":
        return HitlResumeRequest(
            checkpoint_id=self.checkpoint_id,
            review_kind=self.kind,
            decisions=[item.approve() for item in self.items],
        )

    def with_decisions(self, *decisions: HitlResumeDecision) -> "HitlResumeRequest":
        return HitlResumeRequest(
            checkpoint_id=self.checkpoint_id,
            review_kind=self.kind,
            decisions=list(decisions),
        )


class HitlResumeRequest(BaseModel):
    checkpoint_id: str = Field(min_length=1)
    review_kind: Literal["subquestion_review", "query_expansion_review"]
    decisions: list[HitlResumeDecision] = Field(min_length=1)

    @classmethod
    def approve(cls, review: HitlReview, item_id: str) -> "HitlResumeRequest":
        return review.with_decisions(HitlResumeDecision(item_id=item_id, action="approve"))

    @classmethod
    def edit(cls, review: HitlReview, item_id: str, replacement_text: str) -> "HitlResumeRequest":
        return review.with_decisions(
            HitlResumeDecision(item_id=item_id, action="edit", replacement_text=replacement_text)
        )

    @classmethod
    def reject(cls, review: HitlReview, item_id: str) -> "HitlResumeRequest":
        return review.with_decisions(HitlResumeDecision(item_id=item_id, action="reject"))


class RuntimeAgentRunResult(BaseModel):
    status: Literal["completed", "paused"]
    thread_id: str = ""
    checkpoint_id: str | None = None
    review: HitlReview | None = Field(
        default=None,
        validation_alias=AliasChoices("review", "interrupt_payload"),
    )
    response: RuntimeAgentRunResponse | None = None

    @field_validator("review", mode="before")
    @classmethod
    def validate_review(cls, value: Any) -> HitlReview | Any | None:
        if value is None or isinstance(value, HitlReview):
            return value
        if isinstance(value, Mapping):
            return HitlReview.from_interrupt_payload(value)
        return value


class AgentRunStageMetadata(BaseModel):
    stage: str
    status: str
    sub_question: str = ""
    lane_index: int = 0
    lane_total: int = 0
    emitted_at: float | None = None


class RuntimeAgentRunAsyncStartResponse(BaseModel):
    job_id: str
    run_id: str
    thread_id: str = ""
    status: str


class RuntimeAgentRunAsyncStatusResponse(BaseModel):
    job_id: str
    run_id: str = ""
    thread_id: str = ""
    status: str
    message: str = ""
    stage: str = ""
    stages: list[AgentRunStageMetadata] = Field(default_factory=list)
    decomposition_sub_questions: list[str] = Field(default_factory=list)
    sub_question_artifacts: list["SubQuestionArtifacts"] = Field(default_factory=list)
    sub_qa: list[SubQuestionAnswer] = Field(default_factory=list)
    sub_answers: list[SubQuestionAnswer] = Field(default_factory=list)
    output: str = ""
    result: RuntimeAgentRunResponse | None = None
    error: str | None = None
    cancel_requested: bool = False
    review: HitlReview | None = Field(
        default=None,
        validation_alias=AliasChoices("review", "interrupt_payload"),
    )
    checkpoint_id: str | None = None
    started_at: float | None = None
    finished_at: float | None = None
    elapsed_ms: int | None = None

    @field_validator("review", mode="before")
    @classmethod
    def validate_review(cls, value: Any) -> HitlReview | Any | None:
        if value is None or isinstance(value, HitlReview):
            return value
        if isinstance(value, Mapping):
            return HitlReview.from_interrupt_payload(value)
        return value


class RuntimeAgentRunAsyncCancelResponse(BaseModel):
    status: Literal["success"]
    message: str


class RuntimeSubquestionDecision(BaseModel):
    subquestion_id: str = Field(min_length=1)
    action: Literal["approve", "edit", "deny", "skip"]
    edited_text: str | None = None

    @model_validator(mode="after")
    def validate_edit_payload(self) -> "RuntimeSubquestionDecision":
        if self.action == "edit":
            if self.edited_text is None or not self.edited_text.strip():
                raise ValueError("edited_text is required when action='edit'.")
        elif self.edited_text is not None:
            self.edited_text = self.edited_text.strip() or None
        return self


class RuntimeSubquestionResumeEnvelope(BaseModel):
    checkpoint_id: str = Field(min_length=1)
    decisions: list[RuntimeSubquestionDecision] = Field(min_length=1)


class RuntimeQueryExpansionDecision(BaseModel):
    expansion_id: str = Field(min_length=1)
    action: Literal["approve", "edit", "deny", "skip"]
    edited_query: str | None = None

    @model_validator(mode="after")
    def validate_edit_payload(self) -> "RuntimeQueryExpansionDecision":
        if self.action == "edit":
            if self.edited_query is None or not self.edited_query.strip():
                raise ValueError("edited_query is required when action='edit'.")
        elif self.edited_query is not None:
            self.edited_query = self.edited_query.strip() or None
        return self


class RuntimeQueryExpansionResumeEnvelope(BaseModel):
    checkpoint_id: str = Field(min_length=1)
    decisions: list[RuntimeQueryExpansionDecision] = Field(min_length=1)


class RuntimeAgentRunResumeRequest(BaseModel):
    resume: Union[
        bool,
        HitlResumeRequest,
        dict[str, Any],
        RuntimeSubquestionResumeEnvelope,
        RuntimeQueryExpansionResumeEnvelope,
    ] = True

    @field_validator("resume", mode="before")
    @classmethod
    def validate_resume(cls, value: Any) -> Any:
        if isinstance(value, bool):
            return value
        if isinstance(value, HitlResumeRequest):
            return value
        if not isinstance(value, dict):
            raise ValueError(
                "resume must be a boolean, SDK HITL resume request, legacy object payload, or typed decision envelope."
            )
        if "review_kind" in value and "checkpoint_id" in value and "decisions" in value:
            return HitlResumeRequest.model_validate(value)
        if "checkpoint_id" in value or "decisions" in value:
            decisions = value.get("decisions")
            if isinstance(decisions, list):
                for decision in decisions:
                    if isinstance(decision, dict) and (
                        "expansion_id" in decision or "edited_query" in decision
                    ):
                        return RuntimeQueryExpansionResumeEnvelope.model_validate(value)
            return RuntimeSubquestionResumeEnvelope.model_validate(value)
        return value


class RuntimeAgentInfo(BaseModel):
    agent_name: str


class GraphRunMetadata(BaseModel):
    run_id: str
    thread_id: str = ""
    trace_id: str = ""
    correlation_id: str = ""


class CitationSourceRow(BaseModel):
    citation_index: int = Field(ge=1)
    rank: int = Field(ge=1)
    title: str = ""
    source: str = ""
    content: str = ""
    document_id: str = ""
    score: float | None = None


class SubQuestionArtifacts(BaseModel):
    sub_question: str
    expanded_queries: list[str] = Field(default_factory=list)
    retrieved_docs: list[CitationSourceRow] = Field(default_factory=list)
    retrieval_provenance: list[dict[str, Any]] = Field(default_factory=list)
    reranked_docs: list[CitationSourceRow] = Field(default_factory=list)
    sub_answer: str = ""
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)


class AgentGraphState(BaseModel):
    main_question: str
    decomposition_sub_questions: list[str] = Field(default_factory=list)
    sub_question_artifacts: list[SubQuestionArtifacts] = Field(default_factory=list)
    final_answer: str = ""
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)
    run_metadata: GraphRunMetadata
    sub_qa: list[SubQuestionAnswer] = Field(default_factory=list)
    output: str = ""
    stage_snapshots: list["GraphStageSnapshot"] = Field(default_factory=list)


class DecomposeNodeInput(BaseModel):
    main_question: str
    run_metadata: GraphRunMetadata
    initial_search_context: list[dict[str, Any]] = Field(default_factory=list)


class DecomposeNodeOutput(BaseModel):
    decomposition_sub_questions: list[str] = Field(default_factory=list)


class ExpandNodeInput(BaseModel):
    main_question: str
    sub_question: str
    run_metadata: GraphRunMetadata


class ExpandNodeOutput(BaseModel):
    expanded_queries: list[str] = Field(default_factory=list)


class SearchNodeInput(BaseModel):
    sub_question: str
    expanded_queries: list[str] = Field(default_factory=list)
    run_metadata: GraphRunMetadata


class SearchNodeOutput(BaseModel):
    retrieved_docs: list[CitationSourceRow] = Field(default_factory=list)
    retrieval_provenance: list[dict[str, Any]] = Field(default_factory=list)
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)


class RerankNodeInput(BaseModel):
    sub_question: str
    expanded_query: str = ""
    retrieved_docs: list[CitationSourceRow] = Field(default_factory=list)
    run_metadata: GraphRunMetadata


class RerankNodeOutput(BaseModel):
    reranked_docs: list[CitationSourceRow] = Field(default_factory=list)
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)


class AnswerSubquestionNodeInput(BaseModel):
    sub_question: str
    reranked_docs: list[CitationSourceRow] = Field(default_factory=list)
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)
    run_metadata: GraphRunMetadata


class AnswerSubquestionNodeOutput(BaseModel):
    sub_answer: str = ""
    citation_indices_used: list[int] = Field(default_factory=list)
    answerable: bool = False
    verification_reason: str = ""
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)


class SynthesizeFinalNodeInput(BaseModel):
    main_question: str
    sub_qa: list[SubQuestionAnswer] = Field(default_factory=list)
    sub_question_artifacts: list[SubQuestionArtifacts] = Field(default_factory=list)
    run_metadata: GraphRunMetadata


class SynthesizeFinalNodeOutput(BaseModel):
    final_answer: str = ""


class GraphStageSnapshot(BaseModel):
    stage: str
    status: str = "completed"
    sub_question: str = ""
    lane_index: int = 0
    lane_total: int = 0
    decomposition_sub_questions: list[str] = Field(default_factory=list)
    sub_qa: list[SubQuestionAnswer] = Field(default_factory=list)
    sub_question_artifacts: list[SubQuestionArtifacts] = Field(default_factory=list)
    output: str = ""



from __future__ import annotations

from collections.abc import Callable
from typing import Any, Mapping

from agent_search.runtime.graph.builder import build_runtime_graph
from agent_search.runtime.graph.state import RuntimeGraphContext, to_runtime_graph_state
from agent_search.runtime.lifecycle_events import LifecycleEventBuilder, RuntimeLifecycleEvent


def _iter_graph_stream(
    *,
    graph: Any,
    graph_input: Any,
    config: Mapping[str, Any],
) -> tuple[Any, list[tuple[str, Any]]]:
    terminal_state: Any = None
    signals: list[tuple[str, Any]] = []
    for item in graph.stream(
        graph_input,
        config=config,
        stream_mode=["values", "tasks", "updates", "checkpoints"],
    ):
        if isinstance(item, tuple) and len(item) == 2:
            mode, payload = item
        else:
            mode, payload = "values", item
        if mode == "values":
            terminal_state = payload
            continue
        signals.append((str(mode), payload))
    return terminal_state, signals


def _build_execution_config(
    *,
    run_metadata: Any,
    config: Mapping[str, Any] | None,
) -> dict[str, Any]:
    resolved = dict(config or {})
    configurable = dict(resolved.get("configurable") or {})
    configurable.setdefault("thread_id", getattr(run_metadata, "thread_id", ""))
    resolved["configurable"] = configurable
    return resolved


def execute_runtime_graph(
    *,
    context: RuntimeGraphContext,
    run_metadata: Any,
    config: Mapping[str, Any] | None = None,
    lifecycle_callback: Callable[[RuntimeLifecycleEvent], None] | None = None,
    resumed: bool = False,
) -> Any:
    graph = build_runtime_graph(context=context)
    graph_input = to_runtime_graph_state(
        context.payload,
        run_metadata=run_metadata,
        initial_search_context=context.initial_search_context,
    )
    execution_config = _build_execution_config(run_metadata=run_metadata, config=config)
    lifecycle_builder = LifecycleEventBuilder(run_metadata=run_metadata) if lifecycle_callback is not None else None
    try:
        if lifecycle_builder is not None:
            start_event = (
                lifecycle_builder.emit_recovery_started()
                if resumed
                else lifecycle_builder.emit_run_started()
            )
            lifecycle_callback(start_event)
        terminal_state, signals = _iter_graph_stream(graph=graph, graph_input=graph_input, config=execution_config)
        if lifecycle_builder is not None:
            for mode, payload in signals:
                for event in lifecycle_builder.consume_stream_signal(mode, payload):
                    lifecycle_callback(event)
            lifecycle_callback(lifecycle_builder.emit_terminal(status="success"))
        if terminal_state is None:
            return graph.invoke(graph_input, config=execution_config)
        return terminal_state
    except Exception as exc:
        if lifecycle_builder is not None:
            lifecycle_callback(lifecycle_builder.emit_terminal(status="error", error=str(exc)))
        raise


__all__ = ["execute_runtime_graph"]



from __future__ import annotations

import inspect
import logging
import os
import threading
import time
import uuid
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any, Optional

from agent_search.errors import SDKConfigurationError
from agent_search.runtime.execution_identity import resolve_thread_identity
from agent_search.runtime.graph.execution import execute_runtime_graph
from agent_search.runtime.graph.state import RuntimeGraphContext
from agent_search.runtime.lifecycle_events import RuntimeLifecycleEvent
from agent_search.runtime.resume import ResumeTransitionError, ensure_resume_allowed
from agent_search.runtime.state import to_rag_state
from db import SessionLocal
from models import RuntimeCheckpointLink
from schemas import (
    AgentRunStageMetadata,
    GraphStageSnapshot,
    RuntimeAgentRunRequest,
    RuntimeAgentRunResponse,
    SubQuestionAnswer,
    SubQuestionArtifacts,
)
from services.agent_service import build_graph_run_metadata, map_graph_state_to_runtime_response
from agent_search.runtime.runner import run_checkpointed_agent

logger = logging.getLogger(__name__)

_VECTOR_COLLECTION_NAME = os.getenv("VECTOR_COLLECTION_NAME", "agent_search_internal_data")
_EXECUTOR = ThreadPoolExecutor(max_workers=4)
_JOB_LOCK = threading.Lock()


@dataclass
class AgentRunJobStatus:
    job_id: str
    run_id: str
    thread_id: str
    status: str
    query: str = ""
    message: str = ""
    stage: str = "queued"
    stages: list[AgentRunStageMetadata] = field(default_factory=list)
    decomposition_sub_questions: list[str] = field(default_factory=list)
    sub_question_artifacts: list[SubQuestionArtifacts] = field(default_factory=list)
    sub_qa: list[SubQuestionAnswer] = field(default_factory=list)
    output: str = ""
    result: Optional[RuntimeAgentRunResponse] = None
    error: Optional[str] = None
    cancel_requested: bool = False
    interrupt_payload: Any | None = None
    checkpoint_id: str | None = None
    lifecycle_events: list[RuntimeLifecycleEvent] = field(default_factory=list)
    runtime_model: Any | None = field(default=None, repr=False)
    runtime_vector_store: Any | None = field(default=None, repr=False)
    started_at: float = field(default_factory=time.time)
    finished_at: Optional[float] = None


_JOBS: dict[str, AgentRunJobStatus] = {}
_TERMINAL_LIFECYCLE_EVENT_TYPES = frozenset({"run.completed", "run.failed", "run.paused"})


def _call_with_supported_kwargs(func: Any, /, **kwargs: Any) -> Any:
    signature = inspect.signature(func)
    if any(parameter.kind is inspect.Parameter.VAR_KEYWORD for parameter in signature.parameters.values()):
        return func(**kwargs)
    supported_kwargs = {
        key: value
        for key, value in kwargs.items()
        if key in signature.parameters
    }
    return func(**supported_kwargs)


def _event_sequence(event_id: str | None) -> int:
    if event_id is None:
        return 0
    _, _, raw_sequence = event_id.rpartition(":")
    try:
        return int(raw_sequence)
    except ValueError:
        return 0


def append_agent_run_event(job_id: str, event: RuntimeLifecycleEvent) -> None:
    with _JOB_LOCK:
        job = _JOBS.get(job_id)
        if job is None:
            return
        job.lifecycle_events.append(event.model_copy(deep=True))


def list_agent_run_events(job_id: str, *, after_event_id: str | None = None) -> list[RuntimeLifecycleEvent]:
    with _JOB_LOCK:
        job = _JOBS.get(job_id)
        if job is None:
            raise SDKConfigurationError("Job not found.")
        min_sequence = _event_sequence(after_event_id)
        return [event.model_copy(deep=True) for event in job.lifecycle_events if _event_sequence(event.event_id) > min_sequence]


def iter_agent_run_events(
    job_id: str,
    *,
    after_event_id: str | None = None,
    poll_interval: float = 0.05,
):
    last_event_id = after_event_id
    while True:
        events = list_agent_run_events(job_id, after_event_id=last_event_id)
        for event in events:
            last_event_id = event.event_id
            yield event
            if event.event_type in _TERMINAL_LIFECYCLE_EVENT_TYPES:
                return
        terminal_status_reached = False
        with _JOB_LOCK:
            job = _JOBS.get(job_id)
            if job is None:
                return
            terminal_status_reached = job.status in {"success", "error", "cancelled", "paused"}
        if terminal_status_reached:
            trailing_events = list_agent_run_events(job_id, after_event_id=last_event_id)
            for event in trailing_events:
                last_event_id = event.event_id
                yield event
                if event.event_type in _TERMINAL_LIFECYCLE_EVENT_TYPES:
                    return
            return
        time.sleep(poll_interval)


def _persist_job_status(job: AgentRunJobStatus) -> None:
    session = SessionLocal()
    try:
        run = resolve_thread_identity(
            session=session,
            run_id=job.run_id,
            thread_id=job.thread_id,
            status=job.status,
            metadata={"job_id": job.job_id, "stage": job.stage, "query": job.query},
        )
        run.status = job.status
        metadata = dict(run.metadata_json or {})
        metadata.update(
            {
                "job_id": job.job_id,
                "stage": job.stage,
                "message": job.message,
                "query": job.query,
            }
        )
        if job.interrupt_payload is not None:
            metadata["interrupt_payload"] = job.interrupt_payload
        run.metadata_json = metadata
        run.error_message = job.error
        if job.status in {"success", "error", "cancelled"}:
            run.completed_at = datetime.now(timezone.utc)
        if job.checkpoint_id:
            checkpoint_link = (
                session.query(RuntimeCheckpointLink)
                .filter(RuntimeCheckpointLink.run_id == job.run_id)
                .one_or_none()
            )
            if checkpoint_link is None:
                checkpoint_link = RuntimeCheckpointLink(
                    run_id=job.run_id,
                    thread_id=job.thread_id,
                    checkpoint_id=job.checkpoint_id,
                    checkpoint_metadata={"job_id": job.job_id},
                    is_latest=True,
                )
                session.add(checkpoint_link)
            else:
                checkpoint_link.thread_id = job.thread_id
                checkpoint_link.checkpoint_id = job.checkpoint_id
                checkpoint_link.checkpoint_metadata = {"job_id": job.job_id}
                checkpoint_link.is_latest = True
        session.commit()
    finally:
        session.close()


def start_agent_run_job(
    payload: RuntimeAgentRunRequest,
    *,
    model: Any | None = None,
    vector_store: Any | None = None,
) -> AgentRunJobStatus:
    if model is None:
        logger.error("Runtime async job rejected missing model")
        raise SDKConfigurationError("model is required and cannot be None")
    if vector_store is None:
        logger.error("Runtime async job rejected missing vector_store")
        raise SDKConfigurationError("vector_store is required and cannot be None")
    job_id = str(uuid.uuid4())
    run_metadata = build_graph_run_metadata(run_id=job_id, thread_id=payload.thread_id)
    status = AgentRunJobStatus(
        job_id=job_id,
        run_id=run_metadata.run_id,
        thread_id=run_metadata.thread_id,
        status="running",
        query=payload.query,
        message="Run queued.",
        stage="queued",
        runtime_model=model,
        runtime_vector_store=vector_store,
    )
    with _JOB_LOCK:
        _JOBS[job_id] = status
    _persist_job_status(status)
    logger.info(
        "Runtime async job created job_id=%s run_id=%s has_model=%s has_vector_store=%s",
        job_id,
        run_metadata.run_id,
        model is not None,
        vector_store is not None,
    )
    _EXECUTOR.submit(_run_agent_job, job_id, payload, run_metadata.run_id, run_metadata.thread_id, model, vector_store)
    return status


def get_agent_run_job(job_id: str) -> Optional[AgentRunJobStatus]:
    with _JOB_LOCK:
        return _JOBS.get(job_id)


def cancel_agent_run_job(job_id: str) -> bool:
    with _JOB_LOCK:
        job = _JOBS.get(job_id)
        if job is None:
            return False
        if job.status in {"success", "error", "cancelled"}:
            return False
        job.cancel_requested = True
        job.status = "cancelling"
        job.message = "Cancellation requested."
        _persist_job_status(job)
        logger.info("Runtime async job cancel requested job_id=%s run_id=%s", job.job_id, job.run_id)
        return True


def resume_agent_run_job(job_id: str, *, resume: Any = True) -> AgentRunJobStatus:
    with _JOB_LOCK:
        job = _JOBS.get(job_id)
        if job is None:
            raise SDKConfigurationError("Job not found.")
        try:
            ensure_resume_allowed(job.status)
        except ResumeTransitionError as exc:
            raise SDKConfigurationError(str(exc)) from exc
        if job.runtime_model is None or job.runtime_vector_store is None:
            raise SDKConfigurationError("Job cannot be resumed because runtime dependencies are unavailable.")
        job.status = "running"
        job.message = "Resuming from checkpoint."
        job.stage = "resuming"
        job.interrupt_payload = None
        job.finished_at = None
        payload = RuntimeAgentRunRequest(query=job.query, thread_id=job.thread_id)
        model = job.runtime_model
        vector_store = job.runtime_vector_store
    _persist_job_status(job)
    _EXECUTOR.submit(_run_agent_job, job_id, payload, job.run_id, job.thread_id, model, vector_store, resume)
    return job


def _stage_from_snapshot(snapshot: GraphStageSnapshot) -> str:
    if snapshot.stage == "decompose":
        return "subquestions_ready"
    return snapshot.stage


def _run_agent_job(
    job_id: str,
    payload: RuntimeAgentRunRequest,
    run_id: str,
    thread_id: str,
    model: Any | None,
    vector_store: Any | None,
    resume: Any | None = None,
) -> None:
    job = get_agent_run_job(job_id)
    if job is None:
        return
    if job.cancel_requested:
        with _JOB_LOCK:
            job.status = "cancelled"
            job.message = "Cancelled before run started."
            job.finished_at = time.time()
        _persist_job_status(job)
        logger.info("Runtime async job cancelled before start job_id=%s run_id=%s", job_id, run_id)
        return

    logger.info("Runtime async job started job_id=%s run_id=%s query_len=%s", job_id, run_id, len(payload.query))
    with _JOB_LOCK:
        job.stage = "initializing"
        job.message = "Initializing vector store."
    _persist_job_status(job)

    try:
        resolved_vector_store = vector_store
        if resolved_vector_store is None:
            raise SDKConfigurationError("vector_store is required and cannot be None")
        if model is None:
            raise SDKConfigurationError("model is required and cannot be None")
        logger.info(
            "Runtime async job vector store ready source=provided job_id=%s run_id=%s",
            job_id,
            run_id,
        )

        initial_search_context: list[dict[str, Any]] = []
        logger.info(
            "Runtime async job initial context retrieval disabled; proceeding with empty context job_id=%s run_id=%s",
            job_id,
            run_id,
        )

        def on_snapshot(snapshot: GraphStageSnapshot, _state: Any) -> None:
            mapped_stage = _stage_from_snapshot(snapshot)
            stage_metadata = AgentRunStageMetadata(
                stage=mapped_stage,
                status=snapshot.status,
                sub_question=snapshot.sub_question,
                lane_index=snapshot.lane_index,
                lane_total=snapshot.lane_total,
                emitted_at=time.time(),
            )
            with _JOB_LOCK:
                current_job = _JOBS.get(job_id)
                if current_job is None:
                    return
                current_job.stage = mapped_stage
                current_job.message = f"Stage completed: {mapped_stage}"
                current_job.stages.append(stage_metadata)
                current_job.decomposition_sub_questions = list(snapshot.decomposition_sub_questions)
                current_job.sub_question_artifacts = [item.model_copy(deep=True) for item in snapshot.sub_question_artifacts]
                current_job.sub_qa = [item.model_copy(deep=True) for item in snapshot.sub_qa]
                current_job.output = snapshot.output
            logger.info(
                "Runtime async job snapshot job_id=%s run_id=%s stage=%s lane_index=%s lane_total=%s subquestions=%s",
                job_id,
                run_id,
                mapped_stage,
                snapshot.lane_index,
                snapshot.lane_total,
                len(snapshot.decomposition_sub_questions),
            )

        run_metadata = build_graph_run_metadata(run_id=run_id, thread_id=thread_id)
        if resume is None:
            state = _call_with_supported_kwargs(
                execute_runtime_graph,
                context=RuntimeGraphContext(
                    payload=payload,
                    model=model,
                    vector_store=resolved_vector_store,
                    initial_search_context=initial_search_context,
                ),
                run_metadata=run_metadata,
                lifecycle_callback=lambda event: append_agent_run_event(job_id, event),
                snapshot_callback=on_snapshot,
            )
            outcome = type(
                "GraphExecutionOutcome",
                (),
                {
                    "status": "completed",
                    "state": state,
                    "response": map_graph_state_to_runtime_response(state),
                    "interrupt_payload": None,
                    "checkpoint_id": None,
                },
            )()
        else:
            outcome = _call_with_supported_kwargs(
                run_checkpointed_agent,
                payload=payload,
                vector_store=resolved_vector_store,
                model=model,
                run_metadata=run_metadata,
                initial_search_context=initial_search_context,
                lifecycle_callback=lambda event: append_agent_run_event(job_id, event),
                snapshot_callback=on_snapshot,
                resume=resume,
            )
        if outcome.status == "paused":
            with _JOB_LOCK:
                current_job = _JOBS.get(job_id)
                if current_job is None:
                    return
                current_job.status = "paused"
                current_job.message = "Paused and awaiting resume input."
                current_job.stage = "paused"
                current_job.interrupt_payload = outcome.interrupt_payload
                current_job.checkpoint_id = outcome.checkpoint_id
            _persist_job_status(current_job)
            logger.info(
                "Runtime async job paused job_id=%s run_id=%s checkpoint_id=%s",
                job_id,
                run_id,
                outcome.checkpoint_id,
            )
            return
        state = outcome.state
        response = outcome.response
        if state is None or response is None:
            raise SDKConfigurationError("Runtime execution completed without a durable result payload.")
        rag_state = to_rag_state(state)
        with _JOB_LOCK:
            current_job = _JOBS.get(job_id)
            if current_job is None:
                return
            if current_job.cancel_requested:
                current_job.status = "cancelled"
                current_job.message = "Cancelled."
            else:
                current_job.status = "success"
                current_job.message = "Completed."
                current_job.result = response
            current_job.interrupt_payload = None
            current_job.checkpoint_id = outcome.checkpoint_id
            current_job.sub_qa = [item.model_copy(deep=True) for item in response.sub_qa]
            current_job.decomposition_sub_questions = list(rag_state["decomposition_sub_questions"])
            current_job.sub_question_artifacts = [item.model_copy(deep=True) for item in rag_state["sub_question_artifacts"]]
            current_job.output = response.output
            current_job.finished_at = time.time()
        _persist_job_status(current_job)
        logger.info(
            "Runtime async job finished job_id=%s run_id=%s status=%s sub_qa_count=%s output_len=%s",
            job_id,
            run_id,
            current_job.status,
            len(response.sub_qa),
            len(response.output),
        )
    except Exception as exc:  # noqa: BLE001
        logger.exception("Runtime async job failed job_id=%s run_id=%s", job_id, run_id)
        with _JOB_LOCK:
            current_job = _JOBS.get(job_id)
            if current_job is None:
                return
            if current_job.cancel_requested:
                current_job.status = "cancelled"
                current_job.message = "Cancelled."
                current_job.error = None
            else:
                current_job.status = "error"
                current_job.message = "Failed."
                current_job.error = str(exc)
            current_job.finished_at = time.time()
        _persist_job_status(current_job)



from __future__ import annotations

import logging
import os
from typing import Any, Callable

from agent_search.vectorstore.protocol import (
    VectorStoreProtocol,
    assert_vector_store_compatible,
)
from schemas import CitationSourceRow, SearchNodeInput, SearchNodeOutput
from services.vector_store_service import search_documents_for_queries

logger = logging.getLogger(__name__)

_QUERY_LOG_MAX = 200
_SEARCH_NODE_K_FETCH = max(1, int(os.getenv("SEARCH_NODE_K_FETCH", "10")))
_SEARCH_NODE_SCORE_THRESHOLD_RAW = os.getenv("SEARCH_NODE_SCORE_THRESHOLD", "0.0")
try:
    _SEARCH_NODE_SCORE_THRESHOLD = (
        float(_SEARCH_NODE_SCORE_THRESHOLD_RAW) if _SEARCH_NODE_SCORE_THRESHOLD_RAW not in (None, "") else None
    )
except ValueError:
    _SEARCH_NODE_SCORE_THRESHOLD = None
_SEARCH_NODE_MERGED_CAP = max(1, int(os.getenv("SEARCH_NODE_MERGED_CAP", "30")))


def _truncate_query(q: str) -> str:
    return q[: _QUERY_LOG_MAX] + "..." if len(q) > _QUERY_LOG_MAX else q


def _normalize_search_queries(*, sub_question: str, expanded_queries: list[str]) -> list[str]:
    candidates = [sub_question, *expanded_queries]
    normalized: list[str] = []
    seen: set[str] = set()
    for candidate in candidates:
        query = (candidate or "").strip()
        if not query:
            continue
        lowered = query.casefold()
        if lowered in seen:
            continue
        normalized.append(query)
        seen.add(lowered)
    return normalized


def _build_document_identity(
    *,
    document_id: str,
    source: str,
    content: str,
) -> str:
    if document_id:
        return f"document_id:{document_id}"
    normalized_source = source.strip().casefold()
    normalized_content = content.strip()
    return f"source_content:{normalized_source}|{normalized_content}"


def _build_citation_row_from_document(*, document: Any, rank: int) -> CitationSourceRow:
    metadata = document.metadata or {}
    title = str(metadata.get("topic") or metadata.get("title") or metadata.get("wiki_page") or "").strip()
    source = str(metadata.get("wiki_url") or metadata.get("source") or "").strip()
    content = str(getattr(document, "page_content", "") or "").strip()
    document_id = str(getattr(document, "id", "") or "").strip()
    return CitationSourceRow(
        citation_index=rank,
        rank=rank,
        title=title,
        source=source,
        content=content,
        document_id=document_id,
        score=None,
    )


def run_search_node(
    *,
    node_input: SearchNodeInput,
    vector_store: VectorStoreProtocol,
    k_fetch: int | None = None,
    score_threshold: float | None = None,
    merged_cap: int | None = None,
    search_documents_for_queries_fn: Callable[..., dict[str, list[Any]]] = search_documents_for_queries,
    assert_vector_store_compatible_fn: Callable[[Any], VectorStoreProtocol] = assert_vector_store_compatible,
    truncate_query_fn: Callable[[str], str] = _truncate_query,
    default_k_fetch: int = _SEARCH_NODE_K_FETCH,
    default_score_threshold: float | None = _SEARCH_NODE_SCORE_THRESHOLD,
    default_merged_cap: int = _SEARCH_NODE_MERGED_CAP,
) -> SearchNodeOutput:
    compatible_store = assert_vector_store_compatible_fn(vector_store)
    effective_k_fetch = max(1, k_fetch or default_k_fetch)
    effective_score_threshold = default_score_threshold if score_threshold is None else score_threshold
    effective_merged_cap = default_merged_cap if merged_cap is None else merged_cap
    normalized_queries = _normalize_search_queries(
        sub_question=node_input.sub_question,
        expanded_queries=node_input.expanded_queries,
    )
    logger.info(
        "Search node start sub_question=%s expanded_query_count=%s normalized_query_count=%s k_fetch=%s run_id=%s trace_id=%s correlation_id=%s",
        truncate_query_fn(node_input.sub_question),
        len(node_input.expanded_queries),
        len(normalized_queries),
        effective_k_fetch,
        node_input.run_metadata.run_id,
        node_input.run_metadata.trace_id,
        node_input.run_metadata.correlation_id,
    )
    if not normalized_queries:
        logger.warning(
            "Search node skipped; no valid queries sub_question=%s run_id=%s",
            truncate_query_fn(node_input.sub_question),
            node_input.run_metadata.run_id,
        )
        return SearchNodeOutput()

    documents_by_query = search_documents_for_queries_fn(
        vector_store=compatible_store,
        queries=normalized_queries,
        k=effective_k_fetch,
        score_threshold=effective_score_threshold,
    )
    merged_rows: list[CitationSourceRow] = []
    retrieval_provenance: list[dict[str, Any]] = []
    seen_document_identities: dict[str, int] = {}

    def _extract_score(document: Any) -> float | None:
        if document is None:
            return None
        metadata = getattr(document, "metadata", {}) or {}
        score = metadata.get("score")
        return float(score) if isinstance(score, (int, float)) else None

    for query_index, query in enumerate(normalized_queries, start=1):
        docs_for_query = documents_by_query.get(query, [])
        for query_rank, document in enumerate(docs_for_query, start=1):
            row = _build_citation_row_from_document(document=document, rank=len(merged_rows) + 1)
            row.score = _extract_score(document)
            document_identity = _build_document_identity(
                document_id=row.document_id,
                source=row.source,
                content=row.content,
            )
            deduped = document_identity in seen_document_identities
            retrieval_provenance.append(
                {
                    "query": query,
                    "query_index": query_index,
                    "query_rank": query_rank,
                    "document_identity": document_identity,
                    "document_id": row.document_id,
                    "source": row.source,
                    "deduped": deduped,
                }
            )
            if deduped:
                existing_index = seen_document_identities[document_identity]
                existing_row = merged_rows[existing_index]
                existing_score = existing_row.score if existing_row.score is not None else float("-inf")
                candidate_score = row.score if row.score is not None else float("-inf")
                if candidate_score > existing_score:
                    existing_row.score = row.score
                continue
            seen_document_identities[document_identity] = len(merged_rows)
            row.rank = len(merged_rows) + 1
            row.citation_index = len(merged_rows) + 1
            merged_rows.append(row)

    if effective_merged_cap and len(merged_rows) > effective_merged_cap:
        merged_rows = merged_rows[:effective_merged_cap]
        for index, row in enumerate(merged_rows, start=1):
            row.rank = index
            row.citation_index = index

    citation_rows_by_index = {item.citation_index: item for item in merged_rows}
    logger.info(
        "Search node complete sub_question=%s query_count=%s raw_candidates=%s merged_candidates=%s run_id=%s",
        truncate_query_fn(node_input.sub_question),
        len(normalized_queries),
        sum(len(documents_by_query.get(query, [])) for query in normalized_queries),
        len(merged_rows),
        node_input.run_metadata.run_id,
    )
    return SearchNodeOutput(
        retrieved_docs=merged_rows,
        retrieval_provenance=retrieval_provenance,
        citation_rows_by_index=citation_rows_by_index,
    )


__all__ = ["run_search_node"]



from __future__ import annotations

from pydantic import BaseModel, Field, field_validator


class DecompositionPlan(BaseModel):
    """Structured output for decomposition-only LLM call."""

    sub_questions: list[str] = Field(default_factory=list, description="Atomic sub-questions for retrieval.")

    @field_validator("sub_questions", mode="before")
    @classmethod
    def _coerce_sub_questions(cls, value: object) -> list[str]:
        if value is None:
            return []
        if isinstance(value, list):
            items = value
        else:
            items = [value]
        cleaned: list[str] = []
        for item in items:
            text = str(item).strip()
            if text:
                cleaned.append(text)
        return cleaned



from __future__ import annotations

import hashlib
from functools import lru_cache

from langchain_core.embeddings import Embeddings

# Keep in sync with vector column size in models.py / pgvector storage.
EMBEDDING_DIM = 1536


class DeterministicHashEmbeddings(Embeddings):
    """Local deterministic embeddings for scaffold/dev usage."""

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        return [self._embed_text(text) for text in texts]

    def embed_query(self, text: str) -> list[float]:
        return self._embed_text(text)

    def _embed_text(self, text: str) -> list[float]:
        seed = text.encode("utf-8")
        values: list[float] = []
        counter = 0

        # Expand SHA-256 output into a stable EMBEDDING_DIM-length vector.
        while len(values) < EMBEDDING_DIM:
            digest = hashlib.sha256(seed + counter.to_bytes(4, byteorder="big")).digest()
            for i in range(0, len(digest), 4):
                chunk = digest[i : i + 4]
                integer = int.from_bytes(chunk, byteorder="big", signed=False)
                values.append((integer / 0xFFFFFFFF) * 2.0 - 1.0)
                if len(values) == EMBEDDING_DIM:
                    break
            counter += 1
        return values


@lru_cache(maxsize=1)
def get_embedding_model() -> Embeddings:
    return DeterministicHashEmbeddings()



import os

from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base, sessionmaker

DATABASE_URL = os.getenv("DATABASE_URL", "postgresql+psycopg://agent_user:agent_pass@db:5432/agent_search")

engine = create_engine(DATABASE_URL, future=True)
SessionLocal = sessionmaker(bind=engine, autoflush=False, autocommit=False, future=True)
Base = declarative_base()


def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()



from __future__ import annotations

import json
import logging
import os
import re
from typing import Any, Callable

from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI

from schemas import DecomposeNodeInput, DecomposeNodeOutput
from schemas.decomposition import DecompositionPlan

logger = logging.getLogger(__name__)

_QUERY_LOG_MAX = 200
_DECOMPOSITION_ONLY_MODEL = os.getenv("DECOMPOSITION_ONLY_MODEL", os.getenv("RUNTIME_AGENT_MODEL", "gpt-4.1-mini"))
_DECOMPOSITION_ONLY_TEMPERATURE = float(os.getenv("DECOMPOSITION_ONLY_TEMPERATURE", "0"))
_DECOMPOSITION_ONLY_MAX_SUBQUESTIONS_RAW = os.getenv("DECOMPOSITION_ONLY_MAX_SUBQUESTIONS", "8")
_OPENAI_API_KEY = os.getenv("OPENAI_API_KEY", "").strip()
_DECOMPOSITION_ONLY_PROMPT = (
    "You are a decomposition planner for retrieval.\n"
    "Task: break the user question into narrow, atomic sub-questions using the provided retrieval context.\n\n"
    "Rules:\n"
    "- Output only sub-questions; do not answer them.\n"
    "- One concept or entity per sub-question.\n"
    "- Every sub-question must end with '?'.\n"
    "- Prefer entities and concepts from the provided context.\n"
    "- Return valid JSON as an array of strings.\n"
)


def _coerce_decomposition_max_subquestions(raw_value: str) -> int:
    try:
        parsed = int(raw_value)
    except (TypeError, ValueError):
        parsed = 8
    return min(10, max(5, parsed))


_DECOMPOSITION_ONLY_MAX_SUBQUESTIONS = _coerce_decomposition_max_subquestions(
    _DECOMPOSITION_ONLY_MAX_SUBQUESTIONS_RAW
)


def _truncate_query(q: str) -> str:
    return q[: _QUERY_LOG_MAX] + "..." if len(q) > _QUERY_LOG_MAX else q


def _normalize_sub_question(text: str) -> str:
    normalized = (text or "").strip()
    if not normalized:
        return ""
    normalized = normalized.rstrip("?.! ").strip()
    if not normalized:
        return ""
    return f"{normalized}?"


def _parse_decomposition_output(*, raw_output: Any, query: str) -> list[str]:
    fallback_question = _normalize_sub_question(query) or "What is the main question?"
    candidates: list[str] = []

    def _extend_candidates_from_json(value: Any) -> bool:
        if isinstance(value, list):
            for item in value:
                if isinstance(item, str) and item.strip():
                    candidates.append(item.strip())
            return True
        if isinstance(value, dict):
            for key in ("sub_questions", "subquestions", "questions"):
                nested = value.get(key)
                if isinstance(nested, list):
                    for item in nested:
                        if isinstance(item, str) and item.strip():
                            candidates.append(item.strip())
                    return True
        return False

    if isinstance(raw_output, DecompositionPlan):
        candidates = list(raw_output.sub_questions or [])
    elif isinstance(raw_output, list):
        for item in raw_output:
            if isinstance(item, str) and item.strip():
                candidates.append(item.strip())
        if not candidates:
            logger.warning("Decomposition output empty; using fallback question")
            return [fallback_question]
    else:
        text = str(raw_output or "").strip()
        if not text:
            logger.warning("Decomposition output empty; using fallback question")
            return [fallback_question]

        parsed_json: Any | None = None
        json_parse_attempted = False
        try:
            json_parse_attempted = True
            parsed = json.loads(text)
            parsed_json = parsed
            _extend_candidates_from_json(parsed)
        except json.JSONDecodeError:
            pass

        if not candidates and not (
            json_parse_attempted and parsed_json is not None and isinstance(parsed_json, (dict, list))
        ):
            lines = [line.strip() for line in text.splitlines() if line.strip()]
            bullet_prefix = re.compile(r"^(?:[-*]|\d+[.)])\s*")
            for line in lines:
                line = bullet_prefix.sub("", line).strip()
                line = line.strip("\"'")
                if line:
                    candidates.append(line)

    normalized: list[str] = []
    seen: set[str] = set()
    for candidate in candidates:
        sub_question = _normalize_sub_question(candidate)
        if not sub_question:
            continue
        lowered = sub_question.lower()
        if lowered in seen:
            continue
        normalized.append(sub_question)
        seen.add(lowered)

    if normalized:
        if len(normalized) > _DECOMPOSITION_ONLY_MAX_SUBQUESTIONS:
            logger.info(
                "Decomposition output truncated count=%s max=%s",
                len(normalized),
                _DECOMPOSITION_ONLY_MAX_SUBQUESTIONS,
            )
            normalized = normalized[:_DECOMPOSITION_ONLY_MAX_SUBQUESTIONS]
        if len(normalized) < 5:
            logger.info(
                "Decomposition output below target_min count=%s target_min=%s",
                len(normalized),
                5,
            )
        return normalized

    logger.warning("Decomposition output malformed; using fallback question")
    return [fallback_question]


def _build_decomposition_only_input_message(query: str, initial_search_context: list[dict[str, Any]]) -> str:
    serialized_context = json.dumps(initial_search_context, ensure_ascii=True)
    return (
        f"User question:\n{query}\n\n"
        "Initial retrieval context:\n"
        f"{serialized_context}\n"
    )


def _run_decomposition_only_llm_call(
    *,
    query: str,
    initial_search_context: list[dict[str, Any]],
    model: BaseChatModel | None = None,
    callbacks: list[Any] | None = None,
) -> list[str]:
    fallback_question = _normalize_sub_question(query) or f"{query.strip()}?"
    if model is None and not _OPENAI_API_KEY:
        logger.info(
            "Decomposition-only LLM call using fallback; OPENAI_API_KEY is not set model=%s",
            _DECOMPOSITION_ONLY_MODEL,
        )
        return [fallback_question]

    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", _DECOMPOSITION_ONLY_PROMPT),
            ("human", "{input_message}"),
        ]
    )
    try:
        llm = model or ChatOpenAI(
            model=_DECOMPOSITION_ONLY_MODEL,
            temperature=_DECOMPOSITION_ONLY_TEMPERATURE,
        )
        logger.info(
            "Decomposition-only LLM call model selection provided_model=%s default_model=%s",
            model is not None,
            _DECOMPOSITION_ONLY_MODEL,
        )
        chain = prompt | llm.with_structured_output(DecompositionPlan)
        invoke_config = {"callbacks": callbacks} if callbacks else None
        if invoke_config:
            result = chain.invoke(
                {"input_message": _build_decomposition_only_input_message(query, initial_search_context)},
                config=invoke_config,
            )
        else:
            result = chain.invoke(
                {"input_message": _build_decomposition_only_input_message(query, initial_search_context)}
            )
        if isinstance(result, DecompositionPlan) and result.sub_questions:
            return result.sub_questions
        logger.warning(
            "Decomposition-only LLM call returned empty content; using fallback model=%s",
            _DECOMPOSITION_ONLY_MODEL,
        )
    except Exception:
        logger.exception(
            "Decomposition-only LLM call failed; using fallback model=%s",
            _DECOMPOSITION_ONLY_MODEL,
        )
    return [fallback_question]


def run_decomposition_node(
    *,
    node_input: DecomposeNodeInput,
    model: BaseChatModel | None = None,
    timeout_s: int | None = None,
    callbacks: list[Any] | None = None,
    run_llm_call_fn: Callable[..., list[str]] = _run_decomposition_only_llm_call,
    parse_output_fn: Callable[..., list[str]] = _parse_decomposition_output,
    normalize_sub_question_fn: Callable[[str], str] = _normalize_sub_question,
    truncate_query_fn: Callable[[str], str] = _truncate_query,
) -> DecomposeNodeOutput:
    _ = timeout_s
    logger.info(
        "Decomposition node start query=%s context_docs=%s run_id=%s trace_id=%s correlation_id=%s",
        truncate_query_fn(node_input.main_question),
        len(node_input.initial_search_context),
        node_input.run_metadata.run_id,
        node_input.run_metadata.trace_id,
        node_input.run_metadata.correlation_id,
    )
    decomposition_raw_output = run_llm_call_fn(
        query=node_input.main_question,
        initial_search_context=node_input.initial_search_context,
        model=model,
        callbacks=callbacks,
    )

    decomposition_raw_output_preview = json.dumps(decomposition_raw_output, ensure_ascii=True)
    logger.info(
        "Decomposition-only LLM output captured output_length=%s output_preview=%s",
        len(decomposition_raw_output),
        truncate_query_fn(decomposition_raw_output_preview),
    )
    decomposition_sub_questions = parse_output_fn(
        raw_output=decomposition_raw_output,
        query=node_input.main_question,
    )
    logger.info(
        "Decomposition output parsed sub_question_count=%s sub_questions=%s run_id=%s",
        len(decomposition_sub_questions),
        json.dumps(decomposition_sub_questions, ensure_ascii=True),
        node_input.run_metadata.run_id,
    )
    return DecomposeNodeOutput(decomposition_sub_questions=decomposition_sub_questions)


__all__ = ["run_decomposition_node"]



from __future__ import annotations

import inspect
import logging
from typing import Any, Protocol, cast, runtime_checkable

from agent_search.errors import SDKConfigurationError
from langchain_core.documents import Document

logger = logging.getLogger(__name__)


@runtime_checkable
class VectorStoreProtocol(Protocol):
    """Required vector-store contract for SDK retrieval integration.

    Semantics:
    - Must return at most `k` LangChain `Document` objects sorted by descending
      relevance for `query`.
    - `filter` is optional and, when provided, should constrain retrieval.
    - Implementations may expose score-returning APIs in addition to this base
      contract, but `similarity_search` is mandatory.
    """

    def similarity_search(
        self,
        query: str,
        k: int,
        filter: dict[str, Any] | None = None,
    ) -> list[Document]:
        """Return top-k documents for the query."""


def assert_vector_store_compatible(vector_store: Any) -> VectorStoreProtocol:
    """Fail fast when a provided store does not satisfy SDK retrieval contract."""
    if not isinstance(vector_store, VectorStoreProtocol):
        logger.error(
            "Vector store compatibility check failed missing similarity_search type=%s",
            type(vector_store).__name__,
        )
        raise SDKConfigurationError(
            "vector_store must implement similarity_search(query, k, filter=None)."
        )

    method = getattr(vector_store, "similarity_search", None)
    if not callable(method):
        logger.error(
            "Vector store compatibility check failed similarity_search not callable type=%s",
            type(vector_store).__name__,
        )
        raise SDKConfigurationError(
            "vector_store must implement callable similarity_search(query, k, filter=None)."
        )

    try:
        signature = inspect.signature(method)
        signature.bind_partial("contract-probe", 1, filter=None)
    except (TypeError, ValueError) as exc:
        logger.error(
            "Vector store compatibility check failed invalid similarity_search signature type=%s error=%s",
            type(vector_store).__name__,
            exc,
        )
        raise SDKConfigurationError(
            "vector_store similarity_search must accept (query, k, filter=None)."
        ) from exc

    logger.info(
        "Vector store compatibility check passed vector_store_type=%s",
        type(vector_store).__name__,
    )
    return cast(VectorStoreProtocol, vector_store)
